# sage_setup: distribution = sagemath-categories
r"""
Algebraic schemes

An algebraic scheme is defined by a set of polynomials in some
suitable affine or projective coordinates. Possible ambient spaces are

* Affine spaces (:class:`AffineSpace
  <sage.schemes.affine.affine_space.AffineSpace_generic>`),

* Projective spaces (:class:`ProjectiveSpace
  <sage.schemes.projective.projective_space.ProjectiveSpace_ring>`), or

* Toric varieties (:class:`ToricVariety
  <sage.schemes.toric.variety.ToricVariety_field>`).

Note that while projective spaces are of course toric varieties themselves,
they are implemented differently in Sage due to efficiency considerations.
You still can create a projective space as a toric variety if you wish.

In the following, we call the corresponding subschemes affine
algebraic schemes, projective algebraic schemes, or toric algebraic
schemes. In the future other ambient spaces, perhaps by means of
gluing relations, may be introduced.

Generally, polynomials `p_0, p_1, \dots, p_n` define an ideal
`I=\left<p_0, p_1, \dots, p_n\right>`. In the projective and toric case, the
polynomials (and, therefore, the ideal) must be homogeneous. The
associated subscheme `V(I)` of the ambient space is, roughly speaking,
the subset of the ambient space on which all polynomials vanish simultaneously.

.. WARNING::

    You should not construct algebraic scheme objects directly. Instead, use
    ``.subscheme()`` methods of ambient spaces. See below for examples.

EXAMPLES:

We first construct the ambient space, here the affine space `\QQ^2`::

    sage: A2 = AffineSpace(2, QQ, 'x, y')
    sage: A2.coordinate_ring().inject_variables()
    Defining x, y

Now we can write polynomial equations in the variables `x` and `y`. For
example, one equation cuts out a curve (a one-dimensional subscheme)::

    sage: V = A2.subscheme([x^2 + y^2 - 1]); V
    Closed subscheme of Affine Space of dimension 2 over Rational Field defined by:
      x^2 + y^2 - 1
    sage: V.dimension()                                                                 # needs sage.libs.singular
    1

Here is a more complicated example in a projective space::

    sage: P3 = ProjectiveSpace(3, QQ, 'x')
    sage: P3.inject_variables()
    Defining x0, x1, x2, x3
    sage: Q = matrix([[x0, x1, x2], [x1, x2, x3]]).minors(2); Q                         # needs sage.modules
    [-x1^2 + x0*x2, -x1*x2 + x0*x3, -x2^2 + x1*x3]
    sage: twisted_cubic = P3.subscheme(Q); twisted_cubic                                # needs sage.libs.singular sage.modules
    Closed subscheme of Projective Space of dimension 3 over Rational Field defined by:
      -x1^2 + x0*x2,
      -x1*x2 + x0*x3,
      -x2^2 + x1*x3
    sage: twisted_cubic.dimension()                                                     # needs sage.libs.singular sage.modules
    1

Note that there are 3 equations in the 3-dimensional ambient space,
yet the subscheme is 1-dimensional. One can show that it is not
possible to eliminate any of the equations, that is, the twisted cubic
is **not** a complete intersection of two polynomial equations.

Let us look at one affine patch, for example the one where `x_0=1` ::

    sage: patch = twisted_cubic.affine_patch(0); patch                                  # needs sage.libs.singular sage.modules
    Closed subscheme of Affine Space of dimension 3 over Rational Field defined by:
      -x1^2 + x2,
      -x1*x2 + x3,
      -x2^2 + x1*x3
    sage: patch.embedding_morphism()                                                    # needs sage.libs.singular sage.modules
    Scheme morphism:
      From: Closed subscheme of Affine Space of dimension 3 over Rational Field
            defined by: -x1^2 + x2, -x1*x2 + x3, -x2^2 + x1*x3
      To:   Closed subscheme of Projective Space of dimension 3 over Rational Field
            defined by: x1^2 - x0*x2, x1*x2 - x0*x3, x2^2 - x1*x3
      Defn: Defined on coordinates by sending (x1, x2, x3) to (1 : x1 : x2 : x3)


AUTHORS:

- David Kohel, William Stein (2005): initial version

- Andrey Novoseltsev (2010-05-17): subschemes of toric varieties

- Volker Braun (2010-12-24): documentation of schemes and refactoring; added
  coordinate neighborhoods and is_smooth()

- Ben Hutz (2014): subschemes of Cartesian products of projective space

- Ben Hutz (2017): split subschemes types into respective folders
"""

# ****************************************************************************
#       Copyright (C) 2010 Volker Braun <vbraun.name@gmail.com>
#       Copyright (C) 2005 David Kohel <kohel@maths.usyd.edu.au>
#       Copyright (C) 2010 Andrey Novoseltsev <novoselt@gmail.com>
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

import sage.rings.abc

from sage.categories.number_fields import NumberFields

from sage.rings.ideal import Ideal_generic
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import RationalField
from sage.rings.finite_rings.finite_field_base import FiniteField

from sage.misc.latex import latex
from sage.misc.misc import is_iterator

from sage.structure.all import Sequence
from sage.structure.richcmp import richcmp, richcmp_method

from sage.arith.functions import lcm
from sage.arith.misc import gcd

import sage.schemes.affine
from . import ambient_space
from . import scheme


def is_AlgebraicScheme(x):
    """
    Test whether ``x`` is an algebraic scheme.

    INPUT:

    - ``x`` -- anything

    OUTPUT:

    boolean; whether ``x`` is an algebraic scheme, that is, a
    subscheme of an ambient space over a ring defined by polynomial
    equations.

    EXAMPLES::

        sage: A2 = AffineSpace(2, QQ, 'x, y')
        sage: A2.coordinate_ring().inject_variables()
        Defining x, y
        sage: V = A2.subscheme([x^2 + y^2]); V
        Closed subscheme of Affine Space of dimension 2 over Rational Field defined by:
          x^2 + y^2
        sage: from sage.schemes.generic.algebraic_scheme import is_AlgebraicScheme
        sage: is_AlgebraicScheme(V)
        doctest:warning...
        DeprecationWarning: The function is_AlgebraicScheme is deprecated; use 'isinstance(..., AlgebraicScheme)' instead.
        See https://github.com/sagemath/sage/issues/38022 for details.
        True

    Affine space is itself not an algebraic scheme, though the closed
    subscheme defined by no equations is::

        sage: from sage.schemes.generic.algebraic_scheme import is_AlgebraicScheme
        sage: is_AlgebraicScheme(AffineSpace(10, QQ))
        False
        sage: V = AffineSpace(10, QQ).subscheme([]); V
        Closed subscheme of Affine Space of dimension 10 over Rational Field defined by:
          (no polynomials)
        sage: is_AlgebraicScheme(V)
        True

    We create a more complicated closed subscheme::

        sage: A,x = AffineSpace(10, QQ).objgens()
        sage: X = A.subscheme([sum(x)]); X
        Closed subscheme of Affine Space of dimension 10 over Rational Field defined by:
          x0 + x1 + x2 + x3 + x4 + x5 + x6 + x7 + x8 + x9
        sage: is_AlgebraicScheme(X)
        True

    ::

        sage: is_AlgebraicScheme(QQ)
        False
        sage: S = Spec(QQ)
        sage: is_AlgebraicScheme(S)
        False
    """
    from sage.misc.superseded import deprecation
    deprecation(38022, "The function is_AlgebraicScheme is deprecated; use 'isinstance(..., AlgebraicScheme)' instead.")
    return isinstance(x, AlgebraicScheme)


# ****************************************************************************
# A quick overview over the class hierarchy:
#
# class AlgebraicScheme(scheme.Scheme)
#    class AlgebraicScheme_subscheme
#       class AlgebraicScheme_subscheme_affine
#       class AlgebraicScheme_subscheme_projective
#       class AlgebraicScheme_subscheme_toric
#          class AlgebraicScheme_subscheme_affine_toric
#    class AlgebraicScheme_quasi
# ****************************************************************************


class AlgebraicScheme(scheme.Scheme):
    """
    An algebraic scheme presented as a subscheme in an ambient space.

    This is the base class for all algebraic schemes, that is, schemes
    defined by equations in affine, projective, or toric ambient
    spaces.
    """
    def __init__(self, A, category=None):
        """
        TESTS::

            sage: from sage.schemes.generic.algebraic_scheme import AlgebraicScheme
            sage: P = ProjectiveSpace(3, ZZ)
            sage: P.category()
            Category of schemes over Integer Ring
            sage: S = AlgebraicScheme(P); S
            Subscheme of Projective Space of dimension 3 over Integer Ring
            sage: S.category()
            Category of schemes over Integer Ring
        """
        if not isinstance(A, ambient_space.AmbientSpace):
            raise TypeError("A (=%s) must be an ambient space")
        self.__A = A
        self.__divisor_group = {}
        scheme.Scheme.__init__(self, A.base_scheme(), category=category)

    def _latex_(self):
        r"""
        Return a LaTeX representation of this algebraic scheme.

        TESTS::

            sage: from sage.schemes.generic.algebraic_scheme import AlgebraicScheme
            sage: P = ProjectiveSpace(3, ZZ)
            sage: S = AlgebraicScheme(P); S
            Subscheme of Projective Space of dimension 3 over Integer Ring
            sage: S._latex_()
            '\\text{Subscheme of ${\\mathbf P}_{\\Bold{Z}}^{3}$}'
        """
        return r"\text{{Subscheme of ${}$}}".format(latex(self.__A))

    def is_projective(self):
        """
        Return ``True`` if ``self`` is presented as a subscheme of an ambient
        projective space.

        OUTPUT: boolean

        EXAMPLES::

            sage: PP.<x,y,z,w> = ProjectiveSpace(3, QQ)
            sage: f = x^3 + y^3 + z^3 + w^3
            sage: R = f.parent()
            sage: I = [f] + [f.derivative(zz) for zz in PP.gens()]
            sage: V = PP.subscheme(I)
            sage: V.is_projective()
            True
            sage: AA.<x,y,z,w> = AffineSpace(4, QQ)
            sage: V = AA.subscheme(I)
            sage: V.is_projective()
            False

        Note that toric varieties are implemented differently than
        projective spaces. This is why this method returns ``False``
        for toric varieties::

            sage: # needs sage.geometry.polyhedron sage.graphs
            sage: PP.<x,y,z,w> = toric_varieties.P(3)
            sage: V = PP.subscheme(x^3 + y^3 + z^3 + w^3)
            sage: V.is_projective()
            False
        """
        return self.ambient_space().is_projective()

    def coordinate_ring(self):
        """
        Return the coordinate ring of this algebraic scheme.  The
        result is cached.

        OUTPUT:

        The coordinate ring. Usually a polynomial ring, or a quotient
        thereof.

        EXAMPLES::

            sage: P.<x, y, z> = ProjectiveSpace(2, ZZ)
            sage: S = P.subscheme([x - y, x - z])
            sage: S.coordinate_ring()
            Quotient of Multivariate Polynomial Ring in x, y, z over Integer Ring
             by the ideal (x - y, x - z)
        """
        try:
            return self._coordinate_ring
        except AttributeError:
            R = self.__A.coordinate_ring()
            I = self.defining_ideal()
            Q = R.quotient(I)
            self._coordinate_ring = Q
            return Q

    def ambient_space(self):
        """
        Return the ambient space of this algebraic scheme.

        EXAMPLES::

            sage: A.<x, y> = AffineSpace(2, GF(5))
            sage: S = A.subscheme([])
            sage: S.ambient_space()
            Affine Space of dimension 2 over Finite Field of size 5

            sage: P.<x, y, z> = ProjectiveSpace(2, ZZ)
            sage: S = P.subscheme([x - y, x - z])
            sage: S.ambient_space() is P
            True
        """
        return self.__A

    def identity_morphism(self):
        """
        Return the identity morphism.

        OUTPUT: the identity morphism of the scheme ``self``

        EXAMPLES::

            sage: X = Spec(QQ)
            sage: X.identity_morphism()
            Scheme endomorphism of Spectrum of Rational Field
              Defn: Identity map
        """
        from sage.schemes.generic.morphism import SchemeMorphism_polynomial_id
        return SchemeMorphism_polynomial_id(self)

    def embedding_morphism(self):
        r"""
        Return the default embedding morphism of ``self``.

        If the scheme `Y` was constructed as a neighbourhood of a
        point `p \in X`, then :meth:`embedding_morphism` returns a
        local isomorphism `f:Y\to X` around the preimage point
        `f^{-1}(p)`. The latter is returned by
        :meth:`embedding_center`.

        If the algebraic scheme `Y` was not constructed as a
        neighbourhood of a point, then the embedding in its
        :meth:`ambient_space` is returned.

        OUTPUT:

        A scheme morphism whose
        :meth:`~morphism.SchemeMorphism.domain` is ``self``.

        * By default, it is the tautological embedding into its own
          ambient space :meth:`ambient_space`.

        * If the algebraic scheme (which itself is a subscheme of an
          auxiliary :meth:`ambient_space`) was constructed as a patch
          or neighborhood of a point then the embedding is the
          embedding into the original scheme.

        * A :exc:`NotImplementedError` is raised if the construction of
          the embedding morphism is not implemented yet.

        EXAMPLES::

            sage: A2.<x,y> = AffineSpace(QQ, 2)
            sage: C = A2.subscheme(x^2 + y^2 - 1)
            sage: C.embedding_morphism()                                                # needs sage.libs.singular
              Scheme morphism:
              From: Closed subscheme of Affine Space of dimension 2 over Rational Field
                    defined by: x^2 + y^2 - 1
              To:   Affine Space of dimension 2 over Rational Field
              Defn: Defined on coordinates by sending (x, y) to (x, y)

            sage: # needs sage.geometry.polyhedron sage.graphs sage.libs.singular
            sage: P1xP1.<x,y,u,v> = toric_varieties.P1xP1()
            sage: P1 = P1xP1.subscheme(x - y)
            sage: P1.embedding_morphism()
            Scheme morphism:
              From: Closed subscheme of 2-d CPR-Fano toric variety covered
                    by 4 affine patches defined by: x - y
              To:   2-d CPR-Fano toric variety covered by 4 affine patches
              Defn: Defined on coordinates by sending [x : y : u : v] to [y : y : u : v]

        So far, the embedding was just in the own ambient space. Now a
        bit more interesting examples::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme((x^2-y^2)*z)
            sage: p = (1,1,0)
            sage: nbhd = X.neighborhood(p); nbhd
            Closed subscheme of Affine Space of dimension 2 over Rational Field defined by:
              -y^2*z - 2*y*z

        Note that `p=(1,1,0)` is a singular point of `X`. So the
        neighborhood of `p` is not just affine space. The
        :meth:`neighborhood` method returns a presentation of
        the neighborhood as a subscheme of an auxiliary 2-dimensional
        affine space::

            sage: nbhd.ambient_space()
            Affine Space of dimension 2 over Rational Field

        But its :meth:`embedding_morphism` is not into this auxiliary
        affine space, but the original subscheme `X`::

            sage: nbhd.embedding_morphism()
            Scheme morphism:
              From: Closed subscheme of Affine Space of dimension 2 over Rational Field
                    defined by: -y^2*z - 2*y*z
              To:   Closed subscheme of Projective Space of dimension 2 over Rational Field
                    defined by: x^2*z - y^2*z
              Defn: Defined on coordinates by sending (y, z) to (1 : y + 1 : z)

        A couple more examples::

            sage: # needs sage.geometry.polyhedron sage.graphs sage.libs.singular
            sage: patch1 = P1xP1.affine_patch(1); patch1
            2-d affine toric variety
            sage: patch1.embedding_morphism()
            Scheme morphism:
              From: 2-d affine toric variety
              To:   2-d CPR-Fano toric variety covered by 4 affine patches
              Defn: Defined on coordinates by sending [y : u] to [1 : y : u : 1]
            sage: subpatch = P1.affine_patch(1); subpatch
            Closed subscheme of 2-d affine toric variety defined by:
              -y + 1
            sage: subpatch.embedding_morphism()
            Scheme morphism:
              From: Closed subscheme of 2-d affine toric variety defined by: -y + 1
              To:   Closed subscheme of 2-d CPR-Fano toric variety covered
                    by 4 affine patches defined by: x - y
              Defn: Defined on coordinates by sending [y : u] to [1 : y : u : 1]
        """
        if '_embedding_morphism' in self.__dict__:
            hom = self._embedding_morphism
            if isinstance(hom, tuple):
                raise hom[0]
            return hom
        ambient = self.ambient_space()
        return self.hom(self.coordinate_ring().gens(), ambient)

    def embedding_center(self):
        r"""
        Return the distinguished point, if there is any.

        If the scheme `Y` was constructed as a neighbourhood of a
        point `p \in X`, then :meth:`embedding_morphism` returns a
        local isomorphism `f:Y\to X` around the preimage point
        `f^{-1}(p)`. The latter is returned by
        :meth:`embedding_center`.

        OUTPUT:

        A point of ``self``. This raises :exc:`AttributeError` if there
        is no distinguished point, depending on how ``self`` was constructed.

        EXAMPLES::

            sage: P3.<w,x,y,z> = ProjectiveSpace(QQ, 3)
            sage: X = P3.subscheme( (w^2-x^2)*(y^2-z^2) )
            sage: p = [1,-1,3,4]
            sage: nbhd = X.neighborhood(p); nbhd
            Closed subscheme of Affine Space of dimension 3 over Rational Field defined by:
              w^2*y^2 - x^2*y^2 + 6*w^2*y - 6*x^2*y + 2*w*y^2 +
              2*x*y^2 - 7*w^2 + 7*x^2 + 12*w*y + 12*x*y - 14*w - 14*x
            sage: nbhd.embedding_center()
            (0, 0, 0)
            sage: nbhd.embedding_morphism()(nbhd.embedding_center())
            (1/4 : -1/4 : 3/4 : 1)
            sage: nbhd.embedding_morphism()
            Scheme morphism:
              From: Closed subscheme of Affine Space of dimension 3 over Rational Field
                    defined by: w^2*y^2 - x^2*y^2 + 6*w^2*y - 6*x^2*y + 2*w*y^2 + 2*x*y^2
                                - 7*w^2 + 7*x^2 + 12*w*y + 12*x*y - 14*w - 14*x
              To:   Closed subscheme of Projective Space of dimension 3 over Rational Field
                    defined by: w^2*y^2 - x^2*y^2 - w^2*z^2 + x^2*z^2
              Defn: Defined on coordinates by sending (w, x, y) to
                    (w + 1 : x - 1 : y + 3 : 4)
        """
        if '_embedding_center' in self.__dict__:
            return self._embedding_center
        raise AttributeError('This algebraic scheme does not have a designated point.')

    def ngens(self):
        """
        Return the number of generators of the ambient space of this
        algebraic scheme.

        EXAMPLES::

            sage: A.<x, y> = AffineSpace(2, GF(5))
            sage: S = A.subscheme([])
            sage: S.ngens()
            2
            sage: P.<x, y, z> = ProjectiveSpace(2, ZZ)
            sage: S = P.subscheme([x - y, x - z])
            sage: P.ngens()
            3
        """
        return self.__A.ngens()

    def _repr_(self):
        """
        Return a string representation of this algebraic scheme.

        TESTS::

            sage: from sage.schemes.generic.algebraic_scheme import AlgebraicScheme
            sage: P = ProjectiveSpace(3, ZZ)
            sage: S = AlgebraicScheme(P); S
            Subscheme of Projective Space of dimension 3 over Integer Ring
            sage: S._repr_()
            'Subscheme of Projective Space of dimension 3 over Integer Ring'
        """
        return "Subscheme of %s" % self.__A

    def _homset(self, *args, **kwds):
        """
        Construct the Hom-set.

        INPUT:

        Same as :class:`sage.schemes.generic.homset.SchemeHomset_generic`.

        OUTPUT: the Hom-set of the ambient space

        EXAMPLES::

            sage: # needs sage.geometry.polyhedron sage.graphs
            sage: P1.<x,y> = toric_varieties.P1()
            sage: type(P1.Hom(P1))
            <class 'sage.schemes.toric.homset.SchemeHomset_toric_variety_with_category'>
            sage: X = P1.subscheme(x - y)
            sage: type(X.Hom(X))
            <class 'sage.schemes.toric.homset.SchemeHomset_toric_variety_with_category'>

        ::

            sage: # needs sage.geometry.polyhedron sage.graphs
            sage: P1xP1 = toric_varieties.P1xP1()
            sage: P1 = toric_varieties.P1()
            sage: P1xP1._homset(P1xP1, P1)
            Set of morphisms
              From: 2-d CPR-Fano toric variety covered by 4 affine patches
              To:   1-d CPR-Fano toric variety covered by 2 affine patches
        """
        return self.__A._homset(*args, **kwds)

    def _point_homset(self, *args, **kwds):
        """
        Construct a point Hom-set. For internal use only.

        TESTS::

            sage: P2.<x,y,z> = ProjectiveSpace(2, ZZ)
            sage: P2._point_homset(Spec(ZZ), P2)
            Set of rational points of Projective Space of dimension 2 over Integer Ring
        """
        return self.__A._point_homset(*args, **kwds)

    def _point(self, *args, **kwds):
        r"""
        Construct a point of ``self``. For internal use only.

        TESTS::

            sage: P2.<x,y,z> = ProjectiveSpace(2, QQ)
            sage: point_homset = P2._point_homset(Spec(QQ), P2)
            sage: P2._point(point_homset, [1,2,1])
            (1 : 2 : 1)
        """
        return self.__A._point(*args, **kwds)


class AlgebraicScheme_quasi(AlgebraicScheme):
    """
    The quasi-affine or quasi-projective scheme `X - Y`, where `X` and `Y`
    are both closed subschemes of a common ambient affine or projective
    space.

    .. WARNING::

        You should not create objects of this class directly. The
        preferred method to construct such subschemes is to use
        :meth:`complement` method of algebraic schemes.

    OUTPUT: an instance of :class:`AlgebraicScheme_quasi`

    EXAMPLES::

        sage: P.<x, y, z> = ProjectiveSpace(2, ZZ)
        sage: S = P.subscheme([])
        sage: T = P.subscheme([x - y])
        sage: T.complement(S)
        Quasi-projective subscheme X - Y
         of Projective Space of dimension 2 over Integer Ring,
         where X is defined by:
           (no polynomials)
         and Y is defined by:
           x - y
    """

    def __init__(self, X, Y):
        """
        The constructor.

        INPUT:

        - ``X``, ``Y`` -- two subschemes of the same ambient space

        TESTS::

            sage: P.<x, y, z> = ProjectiveSpace(2, ZZ)
            sage: S = P.subscheme([])
            sage: T = P.subscheme([x - y])
            sage: from sage.schemes.generic.algebraic_scheme import AlgebraicScheme_quasi
            sage: AlgebraicScheme_quasi(S, T)
            Quasi-projective subscheme X - Y of Projective Space of dimension 2 over Integer Ring,
             where X is defined by:
               (no polynomials)
             and Y is defined by:
               x - y
        """
        self.__X = X
        self.__Y = Y
        if not isinstance(X, AlgebraicScheme_subscheme):
            raise TypeError("X must be a closed subscheme of an ambient space.")
        if not isinstance(Y, AlgebraicScheme_subscheme):
            raise TypeError("Y must be a closed subscheme of an ambient space.")
        if X.ambient_space() != Y.ambient_space():
            raise ValueError("X and Y must be embedded in the same ambient space.")
        # _latex_ and _repr_ assume all of the above conditions and should be
        # probably changed if they are relaxed!
        A = X.ambient_space()
        self._base_ring = A.base_ring()
        AlgebraicScheme.__init__(self, A)

    def _latex_(self):
        """
        Return a LaTeX representation of this algebraic scheme.

        EXAMPLES::

            sage: from sage.schemes.generic.algebraic_scheme import AlgebraicScheme_quasi
            sage: P.<x, y, z> = ProjectiveSpace(2, ZZ)
            sage: S = P.subscheme([])
            sage: T = P.subscheme([x - y])
            sage: U = AlgebraicScheme_quasi(S, T); U
            Quasi-projective subscheme X - Y of Projective Space of dimension 2 over Integer Ring,
             where X is defined by:
               (no polynomials)
             and Y is defined by:
               x - y
            sage: U._latex_()
            '\\text{Quasi-projective subscheme }
             (X\\setminus Y)\\subset {\\mathbf P}_{\\Bold{Z}}^{2},\\text{ where }
             X \\text{ is defined by }\\text{no polynomials},\\text{ and }
             Y \\text{ is defined by } x - y.'
        """
        if isinstance(self.ambient_space(), sage.schemes.affine.affine_space.AffineSpace_generic):
            t = "affine"
        else:
            t = "projective"
        X = ', '.join(latex(f) for f in self.__X.defining_polynomials())
        if not X:
            X = r"\text{no polynomials}"
        Y = ', '.join(latex(f) for f in self.__Y.defining_polynomials())
        if not Y:
            Y = r"\text{no polynomials}"
        return (r"\text{Quasi-%s subscheme } (X\setminus Y)\subset %s,"
                r"\text{ where } X \text{ is defined by }%s,"
                r"\text{ and } Y \text{ is defined by } %s."
                % (t, latex(self.ambient_space()), X, Y))

    def _repr_(self):
        r"""
        Return a string representation of this algebraic scheme.

        EXAMPLES::

            sage: from sage.schemes.generic.algebraic_scheme import AlgebraicScheme_quasi
            sage: P.<x, y, z> = ProjectiveSpace(2, ZZ)
            sage: S = P.subscheme([])
            sage: T = P.subscheme([x - y])
            sage: U = AlgebraicScheme_quasi(S, T); U
            Quasi-projective subscheme X - Y of Projective Space of dimension 2 over Integer Ring,
             where X is defined by:
               (no polynomials)
             and Y is defined by:
               x - y
            sage: U._repr_()
            'Quasi-projective subscheme X - Y of Projective Space of dimension 2 over Integer Ring, where X is defined by:\n  (no polynomials)\nand Y is defined by:\n  x - y'
        """
        if isinstance(self.ambient_space(), sage.schemes.affine.affine_space.AffineSpace_generic):
            t = "affine"
        else:
            t = "projective"
        return ("Quasi-%s subscheme X - Y of %s, where X is defined by:\n%s\n"
                "and Y is defined by:\n%s"
                % (t, self.ambient_space(), str(self.__X).split("\n", 1)[1],
                   str(self.__Y).split("\n", 1)[1]))

    def X(self):
        """
        Return the scheme `X` such that ``self`` is represented as `X - Y`.

        EXAMPLES::

            sage: P.<x, y, z> = ProjectiveSpace(2, ZZ)
            sage: S = P.subscheme([])
            sage: T = P.subscheme([x - y])
            sage: U = T.complement(S)
            sage: U.X() is S
            True
        """
        return self.__X

    def Y(self):
        """
        Return the scheme `Y` such that ``self`` is represented as `X - Y`.

        EXAMPLES::

            sage: P.<x, y, z> = ProjectiveSpace(2, ZZ)
            sage: S = P.subscheme([])
            sage: T = P.subscheme([x - y])
            sage: U = T.complement(S)
            sage: U.Y() is T
            True
        """
        return self.__Y

    def _check_satisfies_equations(self, v):
        """
        Verify that the coordinates of v define a point on this scheme, or
        raise a :exc:`TypeError`.

        EXAMPLES::

            sage: P.<x, y, z> = ProjectiveSpace(2, ZZ)
            sage: S = P.subscheme([])
            sage: T = P.subscheme([x - y])
            sage: U = T.complement(S)
            sage: U._check_satisfies_equations([1, 2, 0])
            True
            sage: U._check_satisfies_equations([1, 1, 0])
            Traceback (most recent call last):
            ...
            TypeError: Coordinates [1, 1, 0] do not define a point on
            Quasi-projective subscheme X - Y of Projective Space of dimension 2 over Integer Ring,
             where X is defined by:
               (no polynomials)
             and Y is defined by:
               x - y

            sage: U._check_satisfies_equations([1, 4])
            Traceback (most recent call last):
            ...
            TypeError: number of arguments does not match number of variables in parent

            sage: A.<x, y> = AffineSpace(2, GF(7))
            sage: S = A.subscheme([x^2 - y])
            sage: T = A.subscheme([x - y])
            sage: U = T.complement(S)
            sage: U._check_satisfies_equations([2, 4])
            True
            sage: U.point([2,4])
            (2, 4)
            sage: U._check_satisfies_equations(_)
            True
            sage: U._check_satisfies_equations([1, 1])
            Traceback (most recent call last):
            ...
            TypeError: Coordinates [1, 1] do not define a point on Quasi-affine
            subscheme X - Y of Affine Space of dimension 2 over Finite
            Field of size 7, where X is defined by:
              x^2 - y
            and Y is defined by:
              x - y
            sage: U._check_satisfies_equations([1, 0])
            Traceback (most recent call last):
            ...
            TypeError: Coordinates [1, 0] do not define a point on Quasi-affine
            subscheme X - Y of Affine Space of dimension 2 over Finite
            Field of size 7, where X is defined by:
              x^2 - y
            and Y is defined by:
              x - y

        TESTS:

        The bug reported at :issue:`12211` has been fixed::

            sage: P.<x, y, z, w> = ProjectiveSpace(3, QQ)
            sage: S = P.subscheme([x])
            sage: T = P.subscheme([y, z])
            sage: U = T.complement(S)
            sage: U._check_satisfies_equations([0, 0, 1, 1])
            True
        """
        coords = list(v)
        for f in self.__X.defining_polynomials():
            if f(coords) != 0:
                raise TypeError("Coordinates %s do not define a point on %s" % (v,self))
        for f in self.__Y.defining_polynomials():
            if f(coords) != 0:
                return True
        raise TypeError("Coordinates %s do not define a point on %s" % (v,self))

    def rational_points(self, **kwds):
        """
        Return the set of rational points on this algebraic scheme
        over the field `F`.

        INPUT: keyword arguments:

        - ``bound`` -- integer (default: 0); the bound for the coordinates for
          subschemes with dimension at least 1

        - ``F`` -- field (default: base ring); the field to compute
          the rational points over

        EXAMPLES::

            sage: A.<x, y> = AffineSpace(2, GF(7))
            sage: S = A.subscheme([x^2 - y])
            sage: T = A.subscheme([x - y])
            sage: U = T.complement(S)
            sage: U.rational_points()
            [(2, 4), (3, 2), (4, 2), (5, 4), (6, 1)]
            sage: U.rational_points(F=GF(7^2, 'b'))                                     # needs sage.rings.finite_rings
            [(2, 4), (3, 2), (4, 2), (5, 4), (6, 1), (b, b + 4), (b + 1, 3*b + 5),
             (b + 2, 5*b + 1), (b + 3, 6), (b + 4, 2*b + 6), (b + 5, 4*b + 1),
             (b + 6, 6*b + 5), (2*b, 4*b + 2), (2*b + 1, b + 3), (2*b + 2, 5*b + 6),
             (2*b + 3, 2*b + 4), (2*b + 4, 6*b + 4), (2*b + 5, 3*b + 6), (2*b + 6, 3),
             (3*b, 2*b + 1), (3*b + 1, b + 2), (3*b + 2, 5), (3*b + 3, 6*b + 3),
             (3*b + 4, 5*b + 3), (3*b + 5, 4*b + 5), (3*b + 6, 3*b + 2),
             (4*b, 2*b + 1), (4*b + 1, 3*b + 2), (4*b + 2, 4*b + 5),
             (4*b + 3, 5*b + 3), (4*b + 4, 6*b + 3), (4*b + 5, 5), (4*b + 6, b + 2),
             (5*b, 4*b + 2), (5*b + 1, 3), (5*b + 2, 3*b + 6), (5*b + 3, 6*b + 4),
             (5*b + 4, 2*b + 4), (5*b + 5, 5*b + 6), (5*b + 6, b + 3), (6*b, b + 4),
             (6*b + 1, 6*b + 5), (6*b + 2, 4*b + 1), (6*b + 3, 2*b + 6), (6*b + 4, 6),
             (6*b + 5, 5*b + 1), (6*b + 6, 3*b + 5)]
        """
        F = kwds.get('F', None)
        bound = kwds.get('bound', 0)
        if F is None:
            F = self.base_ring()

        if bound == 0:
            if isinstance(F, RationalField):
                raise TypeError("A positive bound (= %s) must be specified." % bound)
            if not isinstance(F, FiniteField):
                raise TypeError("Argument F (= %s) must be a finite field." % F)
        pts = []
        for P in self.ambient_space().rational_points(F):
            try:
                if self._check_satisfies_equations(list(P)):
                    pts.append(P)
            except TypeError:
                pass
        pts.sort()
        return pts


@richcmp_method
class AlgebraicScheme_subscheme(AlgebraicScheme):
    """
    An algebraic scheme presented as a closed subscheme is defined by
    explicit polynomial equations. This is as opposed to a general
    scheme, which could, e.g., be the Neron model of some object, and
    for which we do not want to give explicit equations.

    INPUT:

    - ``A`` -- ambient space (e.g. affine or projective `n`-space)

    - ``polynomials`` -- single polynomial, ideal or iterable of defining
      polynomials; in any case polynomials must belong to the coordinate
      ring of the ambient space and define valid polynomial functions (e.g.
      they should be homogeneous in the case of a projective space)

    OUTPUT: algebraic scheme

    EXAMPLES::

        sage: from sage.schemes.generic.algebraic_scheme import AlgebraicScheme_subscheme
        sage: P.<x, y, z> = ProjectiveSpace(2, QQ)
        sage: P.subscheme([x^2 - y*z])
        Closed subscheme of Projective Space of dimension 2 over Rational Field defined by:
          x^2 - y*z
        sage: AlgebraicScheme_subscheme(P, [x^2 - y*z])
        Closed subscheme of Projective Space of dimension 2 over Rational Field defined by:
          x^2 - y*z
    """

    def __init__(self, A, polynomials, category=None):
        """
        See ``AlgebraicScheme_subscheme`` for documentation.

        TESTS::

            sage: from sage.schemes.generic.algebraic_scheme import AlgebraicScheme_subscheme
            sage: P.<x, y, z> = ProjectiveSpace(2, QQ)
            sage: P.subscheme([x^2 - y*z])
            Closed subscheme of Projective Space of dimension 2 over Rational Field defined by:
              x^2 - y*z
            sage: AlgebraicScheme_subscheme(P, [x^2 - y*z])
            Closed subscheme of Projective Space of dimension 2 over Rational Field defined by:
              x^2 - y*z
        """
        from sage.rings.polynomial.multi_polynomial_sequence import PolynomialSequence_generic

        AlgebraicScheme.__init__(self, A, category=category)
        self._base_ring = A.base_ring()
        R = A.coordinate_ring()
        if isinstance(polynomials, Ideal_generic):
            I = polynomials
            polynomials = I.gens()
            if I.ring() is R:  # Otherwise we will recompute I later after
                self.__I = I  # converting generators to the correct ring
        if isinstance(polynomials, (tuple, PolynomialSequence_generic)) or is_iterator(polynomials):
            polynomials = list(polynomials)
        elif not isinstance(polynomials, list):
            # Looks like we got a single polynomial
            polynomials = [polynomials]
        for n, f in enumerate(polynomials):
            try:
                polynomials[n] = R(f)
            except TypeError:
                raise TypeError("%s cannot be converted to a polynomial in "
                                "the coordinate ring of this %s!" % (f, A))
        polynomials = tuple(polynomials)
        self.__polys = A._validate(polynomials)

    def _check_satisfies_equations(self, v):
        """
        Verify that the coordinates of v define a point on this scheme, or
        raise a :exc:`TypeError`.

        EXAMPLES::

            sage: P.<x, y, z> = ProjectiveSpace(2, QQ)
            sage: S = P.subscheme([x^2 - y*z])
            sage: S._check_satisfies_equations([1, 1, 1])
            True
            sage: S._check_satisfies_equations([1, 0, 1])
            Traceback (most recent call last):
            ...
            TypeError: Coordinates [1, 0, 1] do not define a point on Closed subscheme
            of Projective Space of dimension 2 over Rational Field defined by:
              x^2 - y*z
            sage: S._check_satisfies_equations([0, 0, 0])
            Traceback (most recent call last):
            ...
            TypeError: Coordinates [0, 0, 0] do not define a point on Closed subscheme
            of Projective Space of dimension 2 over Rational Field defined by:
              x^2 - y*z
        """
        coords = list(v)
        for f in self.defining_polynomials():
            if f(coords) != 0:   # it must be "!=0" instead of "if f(v)", e.g.,
                                 # because of p-adic base rings.
                raise TypeError("Coordinates %s do not define a point on %s" % (coords,self))
        try:
            return self.ambient_space()._check_satisfies_equations(coords)
        except TypeError:
            raise TypeError("Coordinates %s do not define a point on %s" % (coords,self))

    def base_extend(self, R):
        """
        Return the base change to the ring `R` of this scheme.

        EXAMPLES::

            sage: P.<x, y, z> = ProjectiveSpace(2, GF(11))
            sage: S = P.subscheme([x^2 - y*z])
            sage: S.base_extend(GF(11^2, 'b'))                                          # needs sage.rings.finite_rings
            Closed subscheme of Projective Space of dimension 2
             over Finite Field in b of size 11^2
             defined by: x^2 - y*z
            sage: S.base_extend(ZZ)
            Traceback (most recent call last):
            ...
            ValueError: no natural map from the base ring (=Finite Field of size 11)
            to R (=Integer Ring)!
        """
        A = self.ambient_space().base_extend(R)
        return A.subscheme(self.__polys)

    def __richcmp__(self, other, op):
        """
        EXAMPLES::

            sage: A.<x, y, z> = AffineSpace(3, QQ)
            sage: X = A.subscheme([x*y, z])
            sage: X == A.subscheme([z, x*y])
            True
            sage: X == A.subscheme([x*y, z^2])                                          # needs sage.libs.singular
            False
            sage: B.<u, v, t> = AffineSpace(3, QQ)
            sage: X == B.subscheme([u*v, t])
            False
        """
        if not isinstance(other, AlgebraicScheme_subscheme):
            return NotImplemented
        A = self.ambient_space()
        if other.ambient_space() != A:
            return NotImplemented
        return richcmp(self.defining_ideal(), other.defining_ideal(), op)

    def _latex_(self):
        """
        Return a LaTeX representation of this scheme.

        EXAMPLES::

            sage: P.<x, y, z> = ProjectiveSpace(2, GF(11))
            sage: S = P.subscheme([x^2 - y*z]); S
            Closed subscheme of Projective Space of dimension 2 over Finite Field of size 11 defined by:
              x^2 - y*z
            sage: S._latex_()
            '\\text{Closed subscheme of } {\\mathbf P}_{\\Bold{F}_{11}}^{2} \\text{ defined by } x^{2} - y z'
            sage: S = P.subscheme([x^2 - y*z, x^5]); S
            Closed subscheme of Projective Space of dimension 2 over Finite Field of size 11 defined by:
              x^2 - y*z,
              x^5
            sage: S._latex_()
            '\\text{Closed subscheme of } {\\mathbf P}_{\\Bold{F}_{11}}^{2} \\text{ defined by } x^{2} - y z, x^{5}'
        """
        polynomials = ', '.join(latex(f) for f in self.defining_polynomials())
        if not polynomials:
            polynomials = r"\text{no polynomials}"
        return (r"\text{Closed subscheme of } %s \text{ defined by } %s"
                % (latex(self.ambient_space()), polynomials))

    def _repr_(self):
        r"""
        Return a string representation of this scheme.

        EXAMPLES::

            sage: P.<x, y, z> = ProjectiveSpace(2, GF(11))
            sage: S = P.subscheme([x^2 - y*z]); S
            Closed subscheme of Projective Space of dimension 2 over Finite Field of size 11 defined by:
              x^2 - y*z
            sage: S._repr_()
            'Closed subscheme of Projective Space of dimension 2 over Finite Field of size 11 defined by:\n  x^2 - y*z'
            sage: S = P.subscheme([x^2 - y*z, x^5]); S
            Closed subscheme of Projective Space of dimension 2 over Finite Field of size 11 defined by:
              x^2 - y*z,
              x^5
            sage: S._repr_()
            'Closed subscheme of Projective Space of dimension 2 over Finite Field of size 11 defined by:\n  x^2 - y*z,\n  x^5'
        """
        polynomials = ',\n  '.join(str(f) for f in self.defining_polynomials())
        if not polynomials:
            polynomials = '(no polynomials)'
        return ("Closed subscheme of %s defined by:\n  %s"
                % (self.ambient_space(), polynomials))

    def defining_polynomials(self):
        """
        Return the polynomials that define this scheme as a subscheme
        of its ambient space.

        OUTPUT:

        A tuple of polynomials in the coordinate ring of the ambient
        space.

        EXAMPLES::

            sage: P.<x, y, z> = ProjectiveSpace(2, ZZ)
            sage: S = P.subscheme([x^2 - y*z, x^3 + z^3])
            sage: S.defining_polynomials()
            (x^2 - y*z, x^3 + z^3)
        """
        return self.__polys

    def normalize_defining_polynomials(self):
        r"""
        Function to normalize the coefficients of defining polynomials
        of given subscheme.

        Normalization as in removing denominator from all the coefficients,
        and then removing any common factor between the coefficients.
        It takes LCM of denominators and then removes common factor among
        coefficients, if any.

        EXAMPLES::

            sage: A.<x,y> = AffineSpace(2, QQ)
            sage: S = A.subscheme([2*x^2 + 4*x*y, 1/8*x + 1/3*y])
            sage: S.normalize_defining_polynomials()
            sage: S.defining_polynomials()
            (x^2 + 2*x*y, 3*x + 8*y)
        """
        BR = self.base_ring()
        if (BR == ZZ
                or isinstance(BR, (sage.rings.abc.AlgebraicField, sage.rings.abc.Order))
                or BR in NumberFields()):
            normalized_polys = []
            initial_polys = list(self.__polys)

            for P in initial_polys:
                # stores value which need to be mutliplied to make all coefficient integers
                mult = lcm([c.denominator() for c in P.coefficients()])
                P = mult*P
                # stores the common factor from all coefficients
                div = gcd(list(P.coefficients()))
                poly_ring = P.parent() # need to coerce, since division might change base ring
                P = poly_ring((BR.one()/div)*P)
                normalized_polys.append(P)

            self.__polys = tuple(normalized_polys)

        else:
                raise NotImplementedError("currently normalization is implemented "
                    "only for QQbar, number fields and number field orders")

    def defining_ideal(self):
        """
        Return the ideal that defines this scheme as a subscheme
        of its ambient space.

        OUTPUT: an ideal in the coordinate ring of the ambient space

        EXAMPLES::

            sage: P.<x, y, z> = ProjectiveSpace(2, ZZ)
            sage: S = P.subscheme([x^2 - y*z, x^3 + z^3])
            sage: S.defining_ideal()
            Ideal (x^2 - y*z, x^3 + z^3) of Multivariate Polynomial Ring in x, y, z
             over Integer Ring
        """
        try:
            return self.__I
        except AttributeError:
            R = self.ambient_space().coordinate_ring()
            self.__I = R.ideal(self.defining_polynomials())
            return self.__I

    # Note: dimension must be implemented by the derived classes
    def codimension(self):
        r"""
        Return the codimension of the algebraic subscheme.

        OUTPUT: integer

        EXAMPLES::

            sage: PP.<x,y,z,w,v> = ProjectiveSpace(4, QQ)
            sage: V = PP.subscheme(x*y)
            sage: V.codimension()                                                       # needs sage.libs.singular
            1
            sage: V.dimension()                                                         # needs sage.libs.singular
            3
        """
        return self.ambient_space().dimension() - self.dimension()

    def irreducible_components(self):
        r"""
        Return the irreducible components of this algebraic scheme, as
        subschemes of the same ambient space.

        OUTPUT:

        an immutable sequence of irreducible subschemes of the ambient
        space of this scheme

        The components are cached.

        EXAMPLES:

        We define what is clearly a union of four hypersurfaces in
        `\P^4_{\QQ}` then find the irreducible components::

            sage: PP.<x,y,z,w,v> = ProjectiveSpace(4, QQ)
            sage: V = PP.subscheme((x^2 - y^2 - z^2) * (w^5 - 2*v^2*z^3) * w * (v^3 - x^2*z))
            sage: V.irreducible_components()                                            # needs sage.libs.singular
            [
            Closed subscheme of Projective Space of dimension 4 over Rational Field defined by:
              w,
            Closed subscheme of Projective Space of dimension 4 over Rational Field defined by:
              x^2 - y^2 - z^2,
            Closed subscheme of Projective Space of dimension 4 over Rational Field defined by:
              x^2*z - v^3,
            Closed subscheme of Projective Space of dimension 4 over Rational Field defined by:
              w^5 - 2*z^3*v^2
            ]

        We verify that the irrelevant ideal is not accidentally returned
        (see :issue:`6920`)::

            sage: PP.<x,y,z,w> = ProjectiveSpace(3, QQ)
            sage: f = x^3 + y^3 + z^3 + w^3
            sage: R = f.parent()
            sage: I = [f] + [f.derivative(zz) for zz in PP.gens()]
            sage: V = PP.subscheme(I)
            sage: V.irreducible_components()                                            # needs sage.libs.singular
            [
            <BLANKLINE>
            ]

        The same polynomial as above defines a scheme with a
        nontrivial irreducible component in affine space (instead of
        the empty scheme as above)::

            sage: AA.<x,y,z,w> = AffineSpace(4, QQ)
            sage: V = AA.subscheme(I)
            sage: V.irreducible_components()                                            # needs sage.libs.singular
            [
            Closed subscheme of Affine Space of dimension 4 over Rational Field defined by:
              w,
              z,
              y,
              x
            ]
        """
        try:
            return self.__irreducible_components
        except AttributeError:
            pass
        I = self.defining_ideal()
        P = I.associated_primes()
        if self.is_projective():
            # In the projective case, we must exclude the prime ideals
            # that contain the irrelevant ideal, which is the ideal
            # generated by the variables, which are the gens of the
            # base ring.
            G = I.ring().gens()
            # We make a list of ideals with the property that "any"
            # of the elements of G are not in the ideal.
            P = [J for J in P if any(g not in J for g in G)]

        A = self.ambient_space()
        C = Sequence([A.subscheme(X) for X in P], check=False, cr=True)
        C.sort(key=lambda scheme: scheme.defining_ideal().gens())
        C.set_immutable()
        self.__irreducible_components = C
        return C

    def is_irreducible(self):
        r"""
        Return whether this subscheme is or is not irreducible.

        OUTPUT: boolean

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: K = QuadraticField(-3)
            sage: P.<x,y,z,w,t,u> = ProjectiveSpace(K, 5)
            sage: X = P.subscheme([x*y - z^2 - K.0*t^2, t*w*x + y*z^2 - u^3])
            sage: X.is_irreducible()                                                    # needs sage.libs.singular
            True

        ::

            sage: P.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P.subscheme([(y + x - z)^2])
            sage: X.is_irreducible()                                                    # needs sage.libs.singular
            False

        ::

            sage: A.<x,y,z,w> = AffineSpace(GF(17), 4)
            sage: X = A.subscheme([
            ....:         x*y*z^2 - x*y*z*w - z*w^2 + w^3,
            ....:         x^3*y*z*w - x*y^3*z - x^2*y*z*w - x^2*w^3 + y^2*w^2 + x*w^3
            ....:     ])
            sage: X.is_irreducible()                                                    # needs sage.libs.singular
            False
        """
        return self.defining_ideal().is_prime()

    def Jacobian_matrix(self):
        r"""
        Return the matrix `\frac{\partial f_i}{\partial x_j}` of
        (formal) partial derivatives.

        OUTPUT: a matrix of polynomials

        EXAMPLES::

            sage: P3.<w,x,y,z> = ProjectiveSpace(3, QQ)
            sage: twisted_cubic = P3.subscheme(matrix([[w, x, y],                       # needs sage.libs.singular
            ....:                                      [x, y, z]]).minors(2))
            sage: twisted_cubic.Jacobian_matrix()                                       # needs sage.libs.singular
            [   y -2*x    w    0]
            [   z   -y   -x    w]
            [   0    z -2*y    x]

        This example addresses issue :issue:`20512`::

            sage: X = P3.subscheme([])
            sage: X.Jacobian_matrix().base_ring() == P3.coordinate_ring()               # needs sage.libs.singular
            True
        """
        R = self.ambient_space().coordinate_ring()
        l = self.defining_polynomials()
        if len(l) == 0:
            return sage.matrix.constructor.Matrix(R, 0)

        from sage.calculus.functions import jacobian

        return jacobian(l, R.gens())

    def Jacobian(self):
        r"""
        Return the Jacobian ideal.

        This is the ideal generated by

        * the `d\times d` minors of the Jacobian matrix, where `d` is
          the :meth:`codimension` of the algebraic scheme, and

        * the defining polynomials of the algebraic scheme. Note that
          some authors do not include these in the definition of the
          Jacobian ideal. An example of a reference that does include
          the defining equations is [Laz2004]_, p. 181.

        OUTPUT: an ideal in the coordinate ring of the ambient space

        EXAMPLES::

            sage: P3.<w,x,y,z> = ProjectiveSpace(3, QQ)
            sage: twisted_cubic = P3.subscheme(matrix([[w, x, y],                       # needs sage.libs.singular
            ....:                                      [x, y, z]]).minors(2))
            sage: twisted_cubic.Jacobian()                                              # needs sage.libs.singular
            Ideal (-x^2 + w*y, -x*y + w*z, -y^2 + x*z, x*z, -2*w*z, w*y, 3*w*y,
                   -2*w*x, w^2, y*z, -2*x*z, w*z, 3*w*z, -2*w*y, w*x, z^2, -2*y*z,
                   x*z, 3*x*z, -2*w*z, w*y)
             of Multivariate Polynomial Ring in w, x, y, z over Rational Field
            sage: twisted_cubic.defining_ideal()                                        # needs sage.libs.singular
            Ideal (-x^2 + w*y, -x*y + w*z, -y^2 + x*z)
             of Multivariate Polynomial Ring in w, x, y, z over Rational Field

        This example addresses issue :issue:`20512`::

            sage: X = P3.subscheme([])
            sage: X.Jacobian() == P3.coordinate_ring().unit_ideal()                     # needs sage.libs.singular
            True
        """
        d = self.codimension()
        minors = self.Jacobian_matrix().minors(d)
        I = self.defining_ideal()
        minors = tuple([ I.reduce(m) for m in minors ])
        return I.ring().ideal(I.gens() + minors)

    def reduce(self):
        r"""
        Return the corresponding reduced algebraic space associated to this
        scheme.

        EXAMPLES: First we construct the union of a doubled and tripled
        line in the affine plane over `\QQ` ::

            sage: A.<x,y> = AffineSpace(2, QQ)
            sage: X = A.subscheme([(x-1)^2*(x-y)^3]); X
            Closed subscheme of Affine Space of dimension 2 over Rational Field defined by:
              x^5 - 3*x^4*y + 3*x^3*y^2 - x^2*y^3 - 2*x^4 + 6*x^3*y
              - 6*x^2*y^2 + 2*x*y^3 + x^3 - 3*x^2*y + 3*x*y^2 - y^3
            sage: X.dimension()                                                         # needs sage.libs.singular
            1

        Then we compute the corresponding reduced scheme::

            sage: Y = X.reduce(); Y                                                     # needs sage.libs.singular
            Closed subscheme of Affine Space of dimension 2 over Rational Field defined by:
              x^2 - x*y - x + y

        Finally, we verify that the reduced scheme `Y` is the union
        of those two lines::

            sage: # needs sage.libs.singular
            sage: L1 = A.subscheme([x - 1]); L1
            Closed subscheme of Affine Space of dimension 2 over Rational Field defined by:
              x - 1
            sage: L2 = A.subscheme([x - y]); L2
            Closed subscheme of Affine Space of dimension 2 over Rational Field defined by:
              x - y
            sage: W = L1.union(L2); W             # taken in ambient space
            Closed subscheme of Affine Space of dimension 2 over Rational Field defined by:
              x^2 - x*y - x + y
            sage: Y == W
            True
        """
        try:
            return self._reduce
        except AttributeError:
            r = self.defining_ideal().radical()
            A = self.ambient_space()
            V = A.subscheme(r)
            V._reduce = V       # so knows it is already reduced!
            self._reduce = V
            return V

    def union(self, other):
        """
        Return the scheme-theoretic union of ``self`` and ``other`` in their common
        ambient space.

        EXAMPLES: We construct the union of a line and a tripled-point on
        the line.

        ::

            sage: A.<x,y> = AffineSpace(2, QQ)
            sage: I = ideal([x, y])^3
            sage: P = A.subscheme(I)
            sage: L = A.subscheme([y - 1])
            sage: S = L.union(P); S                                                     # needs sage.libs.singular
            Closed subscheme of Affine Space of dimension 2 over Rational Field defined by:
              y^4 - y^3,
              x*y^3 - x*y^2,
              x^2*y^2 - x^2*y,
              x^3*y - x^3
            sage: S.dimension()                                                         # needs sage.libs.singular
            1
            sage: S.reduce()                                                            # needs sage.libs.singular
            Closed subscheme of Affine Space of dimension 2 over Rational Field defined by:
              y^2 - y,
              x*y - x

        We can also use the notation "+" for the union::

            sage: A.subscheme([x]) + A.subscheme([y^2 - (x^3+1)])                       # needs sage.libs.singular
            Closed subscheme of Affine Space of dimension 2 over Rational Field defined by:
              x^4 - x*y^2 + x

        Saving and loading::

            sage: loads(S.dumps()) == S                                                 # needs sage.libs.singular
            True
        """
        if not isinstance(other, AlgebraicScheme_subscheme):
            raise TypeError("other (=%s) must be a closed algebraic subscheme of an ambient space" % other)
        A = self.ambient_space()
        if other.ambient_space() != A:
            raise ValueError("other (=%s) must be in the same ambient space as self" % other)
        return A.subscheme(self.defining_ideal().intersection(other.defining_ideal()))

    def __pow__(self, m):
        """
        Return the Cartesian power of this space.

        INPUT:

        - ``m`` -- integer

        OUTPUT: subscheme of product of ambient spaces

        EXAMPLES::

            sage: P2.<y0,y1,y2> = ProjectiveSpace(ZZ, 2)
            sage: Z = P2.subscheme([y0^2 - y1*y2, y2])
            sage: Z**3
            Closed subscheme of Product of projective spaces P^2 x P^2 x P^2 over
             Integer Ring defined by:
              x0^2 - x1*x2,
              x2,
              x3^2 - x4*x5,
              x5,
              x6^2 - x7*x8,
              x8

            sage: A2.<x,y> = AffineSpace(QQ, 2)
            sage: V = A2.subscheme([x^2 - y, x - 1])
            sage: V**4
            Closed subscheme of Affine Space of dimension 8 over Rational Field
             defined by:
              x0^2 - x1,
              x0 - 1,
              x2^2 - x3,
              x2 - 1,
              x4^2 - x5,
              x4 - 1,
              x6^2 - x7,
              x6 - 1

            sage: T.<x0,x1,x2,x3,x4,x5> = ProductProjectiveSpaces([2,2], ZZ)
            sage: X = T.subscheme([x0*x4 - x1*x3])
            sage: X^2
            Closed subscheme of Product of projective spaces P^2 x P^2 x P^2 x P^2
             over Integer Ring defined by:
              -x1*x3 + x0*x4,
              -x7*x9 + x6*x10

            sage: E = EllipticCurve([0,0,0,0,1])                                        # needs sage.schemes
            sage: E^2                                                                   # needs sage.schemes
            Closed subscheme of Product of projective spaces P^2 x P^2
             over Rational Field defined by:
              -x0^3 + x1^2*x2 - x2^3,
              -x3^3 + x4^2*x5 - x5^3
        """
        AS = self.ambient_space().__pow__(m)
        CR = AS.coordinate_ring()
        n = self.ambient_space().coordinate_ring().ngens()

        polys = []
        for i in range(m):
            phi = self.ambient_space().coordinate_ring().hom(list(CR.gens()[n*i : n*(i+1)]), CR)
            polys.extend([phi(t) for t in self.defining_polynomials()])
        return AS.subscheme(polys)

    def __mul__(self, right):
        r"""
        Create the product of subschemes.

        INPUT:

        - ``right`` -- a subscheme of similar type

        OUTPUT: a subscheme of a the product of the ambient spaces

        EXAMPLES::

            sage: S = ProductProjectiveSpaces([1,2,1], ZZ, 't')
            sage: T = ProductProjectiveSpaces([2,2], ZZ, 'x')
            sage: T.inject_variables()
            Defining x0, x1, x2, x3, x4, x5
            sage: X = T.subscheme([x0*x4 - x1*x3])
            sage: X*S
            Closed subscheme of
             Product of projective spaces P^2 x P^2 x P^1 x P^2 x P^1 over Integer Ring
             defined by:
              -x1*x3 + x0*x4

        ::

            sage: S = ProjectiveSpace(ZZ, 2, 't')
            sage: T.<x0,x1,x2,x3> = ProjectiveSpace(ZZ, 3)
            sage: X = T.subscheme([x0*x2 - x1*x3])
            sage: X*S
            Closed subscheme of Product of projective spaces P^3 x P^2 over Integer Ring
             defined by:
              x0*x2 - x1*x3

        ::

            sage: A2 = AffineSpace(ZZ, 2, 't')
            sage: A3.<x0,x1,x2> = AffineSpace(ZZ, 3)
            sage: X = A3.subscheme([x0*x2 - x1])
            sage: X*A2
            Closed subscheme of Affine Space of dimension 5 over Integer Ring
             defined by:
              x0*x2 - x1

        ::

            sage: T.<x0,x1,x2,x3,x4,x5> = ProductProjectiveSpaces([2,2], ZZ)
            sage: X = T.subscheme([x0*x4 - x1*x3])
            sage: X*X
            Closed subscheme of Product of projective spaces P^2 x P^2 x P^2 x P^2
             over Integer Ring defined by:
              -x1*x3 + x0*x4,
              -x7*x9 + x6*x10

        ::

            sage: P1.<z0,z1> = ProjectiveSpace(ZZ, 1)
            sage: Y = P1.subscheme([z0 - z1])
            sage: T.<x0,x1,x2,x3,x4,x5> = ProductProjectiveSpaces([2,2], ZZ)
            sage: X = T.subscheme([x0*x4 - x1*x3])
            sage: X*Y
            Closed subscheme of Product of projective spaces P^2 x P^2 x P^1
             over Integer Ring defined by:
              -x1*x3 + x0*x4,
              z0 - z1

        ::

            sage: A3.<x0,x1,x2> = AffineSpace(ZZ, 3)
            sage: X = A3.subscheme([x0*x2 - x1])
            sage: P1.<u,v> = ProjectiveSpace(ZZ, 1)
            sage: Y = P1.subscheme([u - v])
            sage: X*Y
            Traceback (most recent call last):
            ...
            TypeError: Projective Space of dimension 1 over Integer Ring must be an affine space or affine subscheme
            sage: Y*X
            Traceback (most recent call last):
            ...
            TypeError: Affine Space of dimension 3 over Integer Ring must be a projective space,
            product of projective spaces, or subscheme
            sage: PP.<a,b,c,d> = ProductProjectiveSpaces(ZZ, [1,1])
            sage: Z = PP.subscheme([a*d - b*c])
            sage: X*Z
            Traceback (most recent call last):
            ...
            TypeError: Product of projective spaces P^1 x P^1 over Integer Ring must be an affine space or affine subscheme
            sage: Z*X
            Traceback (most recent call last):
            ...
            TypeError: Affine Space of dimension 3 over Integer Ring must be a projective space,
            product of projective spaces, or subscheme
        """
        #This will catch any ambient space mismatches
        AS = self.ambient_space()*right.ambient_space()
        CR = AS.coordinate_ring()
        n = self.ambient_space().coordinate_ring().ngens()

        phi = self.ambient_space().coordinate_ring().hom(list(CR.gens()[:n]), CR)
        psi = right.ambient_space().coordinate_ring().hom(list(CR.gens()[n:]), CR)
        return AS.subscheme([phi(t) for t in self.defining_polynomials()] + [psi(t) for t in right.defining_polynomials()])

    __add__ = union

    def intersection(self, other):
        """
        Return the scheme-theoretic intersection of ``self`` and ``other`` in their
        common ambient space.

        EXAMPLES::

            sage: A.<x, y> = AffineSpace(2, ZZ)
            sage: X = A.subscheme([x^2 - y])
            sage: Y = A.subscheme([y])
            sage: X.intersection(Y)
            Closed subscheme of Affine Space of dimension 2 over Integer Ring defined by:
              x^2 - y,
              y
        """
        if not isinstance(other, AlgebraicScheme_subscheme):
            raise TypeError("other (=%s) must be a closed algebraic subscheme of an ambient space" % other)
        A = self.ambient_space()
        if other.ambient_space() != A:
            raise ValueError("other (=%s) must be in the same ambient space as self" % other)
        return A.subscheme(self.defining_ideal() + other.defining_ideal())

    def complement(self, other=None):
        """
        Return the scheme-theoretic complement ``other - self``, where
        ``self`` and ``other`` are both closed algebraic subschemes of the
        same ambient space.

        If ``other`` is unspecified, it is taken to be the ambient space
        of ``self``.

        EXAMPLES::

            sage: A.<x, y, z> = AffineSpace(3, ZZ)
            sage: X = A.subscheme([x + y - z])
            sage: Y = A.subscheme([x - y + z])
            sage: Y.complement(X)
            Quasi-affine subscheme X - Y of
             Affine Space of dimension 3 over Integer Ring,
             where X is defined by:
               x + y - z
             and Y is defined by:
               x - y + z
            sage: Y.complement()
            Quasi-affine subscheme X - Y of
             Affine Space of dimension 3 over Integer Ring,
             where X is defined by:
               (no polynomials)
             and Y is defined by:
               x - y + z
            sage: P.<x, y, z> = ProjectiveSpace(2, QQ)
            sage: X = P.subscheme([x^2 + y^2 + z^2])
            sage: Y = P.subscheme([x*y + y*z + z*x])
            sage: Y.complement(X)
            Quasi-projective subscheme X - Y of
             Projective Space of dimension 2 over Rational Field,
             where X is defined by:
               x^2 + y^2 + z^2
             and Y is defined by:
               x*y + x*z + y*z
            sage: Y.complement(P)
            Quasi-projective subscheme X - Y of
             Projective Space of dimension 2 over Rational Field,
             where X is defined by:
               (no polynomials)
             and Y is defined by:
               x*y + x*z + y*z
        """
        A = self.ambient_space()
        if other is None:
            other = A.subscheme([])
        elif not isinstance(other, AlgebraicScheme_subscheme):
            if other == A:
                other = A.subscheme([])
            else:
                raise TypeError("Argument other (=%s) must be a closed algebraic subscheme of an ambient space" % other)
        if other.ambient_space() != A:
            raise ValueError("other (=%s) must be in the same ambient space as self" % other)
        return AlgebraicScheme_quasi(other, self)

    def rational_points(self, **kwds):
        """
        Return the rational points on the algebraic subscheme.

        For a dimension 0 subscheme, if the base ring is a numerical field
        such as the ComplexField the results returned could be very far from correct.
        If the polynomials defining the subscheme are defined over a number field, you
        will get better results calling rational points with `F` defined as the number
        field and the base ring as the field of definition. If the base ring
        is a number field, the embedding into ``F`` must be known.

        In the case of numerically approximated points, the points are returned over as
        points of the ambient space.

        For a dimension greater than 0 scheme, depending on bound size, either the
        points in the ambient space are enumerated or a sieving algorithm lifting points
        modulo primes is used. See the documentation in homset for the details of the
        sieving algorithm.

        INPUT: keyword arguments:

        - ``bound`` -- integer (default: 0); the bound for the coordinates for
          subschemes with dimension at least 1

        - ``prec`` -- integer (default: 53); the precision to use to
          compute the elements of bounded height for number fields

        - ``F`` -- field (default: base ring). The field to compute
          the rational points over

        - ``point_tolerance`` -- positive real number (default: 10^(-10)).
          For numerically inexact fields, two points are considered the same
          if their coordinates are within tolerance.

        - ``zero_tolerance`` -- positive real number (default: 10^(-10)).
          For numerically inexact fields, points are on the subscheme if they
          satisfy the equations to within tolerance.

        - ``tolerance`` -- a rational number in (0,1] used in Doyle-Krumm
          algorithm-4

        OUTPUT: list of points in subscheme or ambient space

        .. WARNING::

           For numerically inexact fields such as ComplexField or RealField the
           list of points returned is very likely to be incomplete at best.

        EXAMPLES:

        Enumerate over a projective scheme over a number field::

            sage: # needs sage.rings.number_field
            sage: u = QQ['u'].0
            sage: K.<v> = NumberField(u^2 + 3)
            sage: A.<x,y> = ProjectiveSpace(K, 1)
            sage: X = A.subscheme(x^2 - y^2)
            sage: X.rational_points(bound=3)                                            # needs sage.libs.singular
            [(-1 : 1), (1 : 1)]

        One can enumerate points up to a given bound on a projective scheme
        over the rationals::

            sage: E = EllipticCurve('37a')                                              # needs sage.schemes
            sage: E.rational_points(bound=8)                                            # needs sage.libs.singular sage.schemes
            [(-1 : -1 : 1), (-1 : 0 : 1), (0 : -1 : 1), (0 : 0 : 1), (0 : 1 : 0),
             (1/4 : -5/8 : 1), (1/4 : -3/8 : 1), (1 : -1 : 1), (1 : 0 : 1),
             (2 : -3 : 1), (2 : 2 : 1)]

        For a small finite field, the complete set of points can be
        enumerated. ::

            sage: Etilde = E.base_extend(GF(3))                                         # needs sage.schemes
            sage: Etilde.rational_points()                                              # needs sage.libs.singular sage.schemes
            [(0 : 1 : 0), (0 : 0 : 1), (0 : 2 : 1), (1 : 0 : 1),
             (1 : 2 : 1), (2 : 0 : 1), (2 : 2 : 1)]

        The class of hyperelliptic curves does not (yet) support
        desingularization of the places at infinity into two points::

            sage: FF = FiniteField(7)
            sage: P.<x> = PolynomialRing(FiniteField(7))
            sage: C = HyperellipticCurve(x^8 + x + 1)                                   # needs sage.schemes
            sage: C.rational_points()                                                   # needs sage.libs.singular sage.schemes
            [(0 : 1 : 0), (0 : 1 : 1), (0 : 6 : 1), (2 : 0 : 1),
             (4 : 0 : 1), (6 : 1 : 1), (6 : 6 : 1)]

        ::

            sage: # needs sage.rings.number_field sage.rings.real_mpfr
            sage: K.<v> = QuadraticField(-3)
            sage: P.<x,y,z> = ProjectiveSpace(K, 2)
            sage: X = P.subscheme([x^2 - v^2*x*z, y*x - v*z^2])
            sage: X.rational_points(F=CC)                                               # needs sage.libs.singular
            [(-3.00000000000000 : -0.577350269189626*I : 1.00000000000000),
             (0.000000000000000 : 1.00000000000000 : 0.000000000000000)]

        ::

            sage: # needs sage.rings.number_field sage.rings.real_mpfr
            sage: K.<v> = QuadraticField(3)
            sage: A.<x,y> = AffineSpace(K, 2)
            sage: X = A.subscheme([x^2 - v^2*y, y*x - v])
            sage: X.rational_points(F=RR)                                               # needs sage.libs.singular
            [(1.73205080756888, 1.00000000000000)]

        .. TODO::

            Implement Stoll's model in weighted projective space to
            resolve singularities and find two points (1 : 1 : 0) and
            (-1 : 1 : 0) at infinity.
        """
        F = kwds.pop('F', None)
        if F is None: #sometimes None is passed in
            F = self.base_ring()
        if F in NumberFields() or F == ZZ:
            X = self.base_extend(F)(F)
            try:
                return X.points(**kwds) # checks for proper bound done in points functions
            except TypeError:
                raise TypeError("Unable to enumerate points over %s." % F)
        elif (self.base_ring() in NumberFields() or self.base_ring() == ZZ)\
          and hasattr(F, 'precision'):
            #we are numerically approximating number field points
            return self(self.base_ring()).numerical_points(F=F, **kwds)
        try:
            X = self.base_extend(F)(F)
            return X.points()
        except TypeError:
            raise TypeError("Unable to enumerate points over %s." % F)

    def change_ring(self, R):
        r"""
        Return a new algebraic subscheme which is this subscheme coerced to ``R``.

        INPUT:

        - ``R`` -- ring or morphism

        OUTPUT: a new algebraic subscheme which is this subscheme coerced to ``R``

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: X = P.subscheme([3*x^2 - y^2])
            sage: H = Hom(X, X)
            sage: X.change_ring(GF(3))
            Closed subscheme of Projective Space of dimension 1
             over Finite Field of size 3 defined by: -y^2

        ::

            sage: # needs sage.rings.number_field
            sage: K.<w> = QuadraticField(2)
            sage: R.<z> = K[]
            sage: L.<v> = K.extension(z^3 - 5)
            sage: P.<x,y> = ProjectiveSpace(K, 1)
            sage: X = P.subscheme(x - w*y)                                              # needs sage.libs.singular
            sage: X.change_ring(L)                                                      # needs sage.libs.singular
            Closed subscheme of Projective Space of dimension 1 over
             Number Field in v with defining polynomial z^3 - 5 over its base field
             defined by: x + (-w)*y

        ::

            sage: # needs sage.rings.number_field
            sage: K.<w> = QuadraticField(2)
            sage: R.<z> = K[]
            sage: L.<v> = K.extension(z^3 - 5)
            sage: P.<x,y,z> = AffineSpace(L, 3)
            sage: X = P.subscheme([x - w*y, z^2 - v*x])                                 # needs sage.libs.singular
            sage: emb = L.embeddings(QQbar)                                             # needs sage.libs.singular
            sage: X.change_ring(emb[0])                                                 # needs sage.libs.singular
            Closed subscheme of Affine Space of dimension 3 over Algebraic Field
             defined by:
              x + (-1.414213562373095? + 0.?e-16*I)*y,
              z^2 + (0.8549879733383485? + 1.480882609682365?*I)*x

        ::

            sage: # needs sage.rings.number_field
            sage: K.<w> = QuadraticField(2)
            sage: R.<z> = K[]
            sage: L.<v> = K.extension(z^3 - 5)
            sage: P.<x,y,z> = AffineSpace(L, 3)
            sage: X = P.subscheme([x - w*y, z^2 - v*x])                                 # needs sage.libs.singular
            sage: emb = L.embeddings(QQbar)                                             # needs sage.libs.singular
            sage: X.change_ring(emb[1])                                                 # needs sage.libs.singular
            Closed subscheme of Affine Space of dimension 3 over Algebraic Field
             defined by:
              x + (-1.414213562373095? + 0.?e-16*I)*y,
              z^2 + (0.8549879733383485? - 1.480882609682365?*I)*x

        ::

            sage: # needs sage.rings.number_field
            sage: K.<w> = QuadraticField(-3)
            sage: P.<x,y> = ProjectiveSpace(K, 1)
            sage: X = P.subscheme(x - w*y)                                              # needs sage.libs.singular
            sage: X.change_ring(CC)                                                     # needs sage.libs.singular
            Closed subscheme of Projective Space of dimension 1
             over Complex Field with 53 bits of precision defined by:
              x + (-1.73205080756888*I)*y

        ::

            sage: # needs sage.rings.number_field
            sage: K.<w> = QuadraticField(3)
            sage: P.<x,y> = ProjectiveSpace(K, 1)
            sage: X = P.subscheme(x - w*y)                                              # needs sage.libs.singular
            sage: X.change_ring(RR)                                                     # needs sage.libs.singular
            Closed subscheme of Projective Space of dimension 1
             over Real Field with 53 bits of precision defined by:
              x - 1.73205080756888*y

        ::

            sage: # needs sage.rings.number_field
            sage: K.<v> = CyclotomicField(7)
            sage: O = K.maximal_order()
            sage: P.<x,y> = ProjectiveSpace(O, 1)
            sage: X = P.subscheme([x^2 + O(v)*y^2])                                     # needs sage.libs.singular
            sage: X.change_ring(CC)                                                     # needs sage.libs.singular
            Closed subscheme of Projective Space of dimension 1
             over Complex Field with 53 bits of precision defined by:
              x^2 + (0.623489801858734 + 0.781831482468030*I)*y^2
            sage: X.change_ring(K).change_ring(K.embeddings(QQbar)[3])                  # needs sage.libs.singular
            Closed subscheme of Projective Space of dimension 1
             over Algebraic Field defined by:
              x^2 + (-0.9009688679024191? - 0.4338837391175581?*I)*y^2

        ::

            sage: # needs sage.rings.number_field
            sage: R.<x> = QQ[]
            sage: f = x^6 - 2
            sage: L.<b> = NumberField(f, embedding=f.roots(CC)[2][0])
            sage: A.<x,y> = AffineSpace(L, 2)
            sage: H = Hom(A, A)
            sage: X = A.subscheme([b*x^2, y^2])                                         # needs sage.libs.singular
            sage: X.change_ring(CC)                                                     # needs sage.libs.singular
            Closed subscheme of Affine Space of dimension 2
             over Complex Field with 53 bits of precision defined by:
              (-0.561231024154687 - 0.972080648619833*I)*x^2,
              y^2
        """
        AS = self.ambient_space()
        new_AS = AS.change_ring(R)
        I = [f.change_ring(R) for f in self.defining_polynomials()]
        return new_AS.subscheme(I)

    def weil_restriction(self):
        r"""
        Compute the Weil restriction of this variety over some extension
        field. If the field is a finite field, then this computes
        the Weil restriction to the prime subfield.

        A Weil restriction of scalars - denoted `Res_{L/k}` - is a
        functor which, for any finite extension of fields `L/k` and
        any algebraic variety `X` over `L`, produces another
        corresponding variety `Res_{L/k}(X)`, defined over `k`. It is
        useful for reducing questions about varieties over large
        fields to questions about more complicated varieties over
        smaller fields.

        This function does not compute this Weil restriction directly
        but computes on generating sets of polynomial ideals:

        Let `d` be the degree of the field extension `L/k`, let `a` a
        generator of `L/k` and `p` the minimal polynomial of
        `L/k`. Denote this ideal by `I`.

        Specifically, this function first maps each variable `x` to
        its representation over `k`: `\sum_{i=0}^{d-1} a^i x_i`. Then
        each generator of `I` is evaluated over these representations
        and reduced modulo the minimal polynomial `p`. The result is
        interpreted as a univariate polynomial in `a` and its
        coefficients are the new generators of the returned ideal.

        If the input and the output ideals are radical, this is
        equivalent to the statement about algebraic varieties above.

        OUTPUT: affine subscheme; the Weil restriction of ``self``

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: R.<x> = QQ[]
            sage: K.<w> = NumberField(x^5 - 2)
            sage: R.<x> = K[]
            sage: L.<v> = K.extension(x^2 + 1)
            sage: A.<x,y> = AffineSpace(L, 2)
            sage: X = A.subscheme([y^2 - L(w)*x^3 - v])                                 # needs sage.libs.singular
            sage: X.weil_restriction()                                                  # needs sage.libs.singular
            Closed subscheme of Affine Space of dimension 4
             over Number Field in w with defining polynomial x^5 - 2 defined by:
              (-w)*z0^3 + (3*w)*z0*z1^2 + z2^2 - z3^2,
              (-3*w)*z0^2*z1 + w*z1^3 + 2*z2*z3 - 1
            sage: X.weil_restriction().ambient_space() is A.weil_restriction()          # needs sage.libs.singular
            True

        ::

            sage: A.<x,y,z> = AffineSpace(GF(5^2, 't'), 3)                              # needs sage.rings.finite_rings
            sage: X = A.subscheme([y^2 - x*z, z^2 + 2*y])                               # needs sage.libs.singular sage.rings.finite_rings
            sage: X.weil_restriction()                                                  # needs sage.libs.singular sage.rings.finite_rings
            Closed subscheme of Affine Space of dimension 6
             over Finite Field of size 5 defined by:
              z2^2 - 2*z3^2 - z0*z4 + 2*z1*z5,
              2*z2*z3 + z3^2 - z1*z4 - z0*z5 - z1*z5,
              z4^2 - 2*z5^2 + 2*z2,
              2*z4*z5 + z5^2 + 2*z3
        """
        try:
            X = self.__weil_restriction
        except AttributeError:
            L = self.base_ring()
            if L.is_finite():
                d = L.degree()
            else:
                d = L.relative_degree()

            if d == 1:
                X = self
            else:
                A = self.ambient_space().weil_restriction()
                I = self.defining_ideal().weil_restriction()
                X = A.subscheme(I)
            self.__weil_restriction = X
        return X

    def specialization(self, D=None, phi=None):
        r"""
        Specialization of this subscheme.

        Given a family of maps defined over a polynomial ring. A specialization
        is a particular member of that family. The specialization can be specified either
        by a dictionary or a :class:`SpecializationMorphism`.

        INPUT:

        - ``D`` -- dictionary (optional)

        - ``phi`` -- :class:`SpecializationMorphism` (optional)

        OUTPUT: :class:`SchemeMorphism_polynomial`

        EXAMPLES::

            sage: R.<c> = PolynomialRing(QQ)
            sage: P.<x,y> = ProjectiveSpace(R, 1)
            sage: X = P.subscheme([x^2 + c*y^2])
            sage: X.specialization(dict({c:2}))
            Closed subscheme of Projective Space of dimension 1 over Rational Field defined by:
              x^2 + 2*y^2

        ::

            sage: R.<c> = PolynomialRing(QQ)
            sage: S.<a,b> = R[]
            sage: P.<x,y,z> = AffineSpace(S, 3)
            sage: X = P.subscheme([x^2 + a*c*y^2 - b*z^2])
            sage: from sage.rings.polynomial.flatten import SpecializationMorphism
            sage: phi = SpecializationMorphism(P.coordinate_ring(),
            ....:                              dict({c: 2, a: 1}))
            sage: X.specialization(phi=phi)                                             # needs sage.libs.singular
            Closed subscheme of Affine Space of dimension 3
             over Univariate Polynomial Ring in b over Rational Field defined by:
              x^2 + 2*y^2 + (-b)*z^2
        """
        if D is None:
            if phi is None:
                raise ValueError("either the dictionary or the specialization must be provided")
        else:
            from sage.rings.polynomial.flatten import SpecializationMorphism
            phi = SpecializationMorphism(self.ambient_space().coordinate_ring(),D)
        amb = self.ambient_space().change_ring(phi.codomain().base_ring())
        return amb.subscheme([phi(g) for g in self.defining_polynomials()])



# sage_setup: distribution = sagemath-categories
r"""
Infinite Polynomial Rings

By Infinite Polynomial Rings, we mean polynomial rings in a countably
infinite number of variables. The implementation consists of a wrapper
around the current *finite* polynomial rings in Sage.

AUTHORS:

- Simon King <simon.king@nuigalway.ie>
- Mike Hansen <mhansen@gmail.com>

An Infinite Polynomial Ring has finitely many generators `x_\ast,
y_\ast,...` and infinitely many variables of the form `x_0, x_1, x_2,
..., y_0, y_1, y_2,...,...`.  We refer to the natural number `n` as
the *index* of the variable `x_n`.

INPUT:

- ``R`` -- the base ring; it has to be a commutative ring, and in some
  applications it must even be a field
- ``names`` -- a finite list of generator names; generator names must be alpha-numeric
- ``order`` -- (optional) string; the default order is ``'lex'`` (lexicographic).
  ``'deglex'`` is degree lexicographic, and ``'degrevlex'`` (degree reverse
  lexicographic) is possible but discouraged.

Each generator ``x`` produces an infinite sequence of variables
``x[1], x[2], ...`` which are printed on screen as ``x_1, x_2, ...``
and are latex typeset as `x_{1}, x_{2}`.  Then, the Infinite
Polynomial Ring is formed by polynomials in these variables.

By default, the monomials are ordered lexicographically. Alternatively,
degree (reverse) lexicographic ordering is possible as well. However, we
do not guarantee that the computation of Groebner bases will terminate
in this case.

In either case, the variables of a Infinite Polynomial Ring X are ordered
according to the following rule:

  ``X.gen(i)[m] > X.gen(j)[n]`` if and only if ``i<j or (i==j and m>n)``

We provide a 'dense' and a 'sparse' implementation. In the dense
implementation, the Infinite Polynomial Ring carries a finite
polynomial ring that comprises *all* variables up to the maximal index
that has been used so far. This is potentially a very big ring and may
also comprise many variables that are not used.

In the sparse implementation, we try to keep the underlying finite
polynomial rings small, using only those variables that are really
needed. By default, we use the dense implementation, since it usually
is much faster.

EXAMPLES::

    sage: X.<x,y> = InfinitePolynomialRing(ZZ, implementation='sparse')
    sage: A.<alpha,beta> = InfinitePolynomialRing(QQ, order='deglex')

    sage: f = x[5] + 2; f
    x_5 + 2
    sage: g = 3*y[1]; g
    3*y_1

It has some advantages to have an underlying ring that is not
univariate.  Hence, we always have at least two variables::

    sage: g._p.parent()
    Multivariate Polynomial Ring in y_1, y_0 over Integer Ring

    sage: f2 = alpha[5] + 2; f2
    alpha_5 + 2
    sage: g2 = 3*beta[1]; g2
    3*beta_1
    sage: A.polynomial_ring()
    Multivariate Polynomial Ring in alpha_5, alpha_4, alpha_3, alpha_2, alpha_1, alpha_0,
     beta_5, beta_4, beta_3, beta_2, beta_1, beta_0 over Rational Field

Of course, we provide the usual polynomial arithmetic::

    sage: f + g
    x_5 + 3*y_1 + 2
    sage: p = x[10]^2*(f+g); p
    x_10^2*x_5 + 3*x_10^2*y_1 + 2*x_10^2
    sage: p2 = alpha[10]^2*(f2+g2); p2
    alpha_10^2*alpha_5 + 3*alpha_10^2*beta_1 + 2*alpha_10^2

There is a permutation action on the variables, by permuting positive
variable indices::

    sage: P = Permutation(((10,1)))
    sage: p^P
    x_5*x_1^2 + 3*x_1^2*y_10 + 2*x_1^2
    sage: p2^P
    alpha_5*alpha_1^2 + 3*alpha_1^2*beta_10 + 2*alpha_1^2

Note that `x_0^P = x_0`, since the permutations only change *positive*
variable indices.

We also implemented ideals of Infinite Polynomial Rings. Here, it is
thoroughly assumed that the ideals are set-wise invariant under the
permutation action. We therefore refer to these ideals as *Symmetric
Ideals*. Symmetric Ideals are finitely generated modulo addition,
multiplication by ring elements and permutation of variables. If the
base ring is a field, one can compute Symmetric Groebner Bases::

    sage: J = A * (alpha[1]*beta[2])
    sage: J.groebner_basis()                                                            # needs sage.combinat sage.libs.singular
    [alpha_1*beta_2, alpha_2*beta_1]

For more details, see :class:`~sage.rings.polynomial.symmetric_ideal.SymmetricIdeal`.

Infinite Polynomial Rings can have any commutative base ring. If the
base ring of an Infinite Polynomial Ring is a (classical or infinite)
Polynomial Ring, then our implementation tries to merge everything
into *one* ring. The basic requirement is that the monomial orders
match. In the case of two Infinite Polynomial Rings, the
implementations must match. Moreover, name conflicts should be
avoided. An overlap is only accepted if the order of variables can be
uniquely inferred, as in the following example::

    sage: A.<a,b,c> = InfinitePolynomialRing(ZZ)
    sage: B.<b,c,d> = InfinitePolynomialRing(A)
    sage: B
    Infinite polynomial ring in a, b, c, d over Integer Ring

This is also allowed if finite polynomial rings are involved::

    sage: A.<a_3,a_1,b_1,c_2,c_0> = ZZ[]
    sage: B.<b,c,d> = InfinitePolynomialRing(A, order='degrevlex')
    sage: B
    Infinite polynomial ring in b, c, d over
     Multivariate Polynomial Ring in a_3, a_1 over Integer Ring

It is no problem if one generator of the Infinite Polynomial Ring is
called ``x`` and one variable of the base ring is also called
``x``. This is since no *variable* of the Infinite Polynomial Ring
will be called ``x``. However, a problem arises if the underlying
classical Polynomial Ring has a variable ``x_1``, since this can be
confused with a variable of the Infinite Polynomial Ring. In this
case, an error will be raised::

    sage: X.<x,y_1> = ZZ[]
    sage: Y.<x,z> = InfinitePolynomialRing(X)

Note that ``X`` is not merged into ``Y``; this is since the monomial
order of ``X`` is 'degrevlex', but of ``Y`` is 'lex'.
::

    sage: Y
    Infinite polynomial ring in x, z over
     Multivariate Polynomial Ring in x, y_1 over Integer Ring

The variable ``x`` of ``X`` can still be interpreted in ``Y``,
although the first generator of ``Y`` is called ``x`` as well::

    sage: x
    x_*
    sage: X('x')
    x
    sage: Y(X('x'))
    x
    sage: Y('x')
    x

But there is only merging if the resulting monomial order is uniquely
determined. This is not the case in the following examples, and thus
an error is raised::

    sage: X.<y_1,x> = ZZ[]
    sage: Y.<y,z> = InfinitePolynomialRing(X)
    Traceback (most recent call last):
    ...
    CoercionException: Overlapping variables (('y', 'z'),['y_1']) are incompatible
    sage: Y.<z,y> = InfinitePolynomialRing(X)
    Traceback (most recent call last):
    ...
    CoercionException: Overlapping variables (('z', 'y'),['y_1']) are incompatible
    sage: X.<x_3,y_1,y_2> = PolynomialRing(ZZ, order='lex')
    sage: # y_1 and y_2 would be in opposite order in an Infinite Polynomial Ring
    sage: Y.<y> = InfinitePolynomialRing(X)
    Traceback (most recent call last):
    ...
    CoercionException: Overlapping variables (('y',),['y_1', 'y_2']) are incompatible


If the type of monomial orderings (e.g., 'degrevlex' versus 'lex') or
if the implementations do not match, there is no simplified
construction available::

    sage: X.<x,y> = InfinitePolynomialRing(ZZ)
    sage: Y.<z> = InfinitePolynomialRing(X, order='degrevlex')
    sage: Y
    Infinite polynomial ring in z over Infinite polynomial ring in x, y over Integer Ring
    sage: Y.<z> = InfinitePolynomialRing(X, implementation='sparse')
    sage: Y
    Infinite polynomial ring in z over Infinite polynomial ring in x, y over Integer Ring

TESTS:

Infinite Polynomial Rings are part of Sage's coercion system. Hence,
we can do arithmetic, so that the result lives in a ring into which
all constituents coerce.
::

    sage: R.<a,b> = InfinitePolynomialRing(ZZ)
    sage: X.<x> = InfinitePolynomialRing(R)
    sage: x[2]/2+(5/3)*a[3]*x[4] + 1
    5/3*a_3*x_4 + 1/2*x_2 + 1

    sage: R.<a,b> = InfinitePolynomialRing(ZZ, implementation='sparse')
    sage: X.<x> = InfinitePolynomialRing(R)
    sage: x[2]/2+(5/3)*a[3]*x[4] + 1
    5/3*a_3*x_4 + 1/2*x_2 + 1

    sage: R.<a,b> = InfinitePolynomialRing(ZZ, implementation='sparse')
    sage: X.<x> = InfinitePolynomialRing(R, implementation='sparse')
    sage: x[2]/2+(5/3)*a[3]*x[4] + 1
    5/3*a_3*x_4 + 1/2*x_2 + 1

    sage: R.<a,b> = InfinitePolynomialRing(ZZ)
    sage: X.<x> = InfinitePolynomialRing(R, implementation='sparse')
    sage: x[2]/2+(5/3)*a[3]*x[4] + 1
    5/3*a_3*x_4 + 1/2*x_2 + 1

Check that :issue:`22514` is fixed::

    sage: R.<x> = InfinitePolynomialRing(ZZ)
    sage: a = R(3)
    sage: a.is_constant()
    True
    sage: a.constant_coefficient()
    3
    sage: a.degree()
    0
    sage: b = R("2")
    sage: b.parent() is R
    True
    sage: S.<y> = ZZ[]
    sage: Q.<z> = InfinitePolynomialRing(S)
    sage: a = Q(1+y)
    sage: a.is_constant()
    True
    sage: a.constant_coefficient()
    y + 1
"""
# ****************************************************************************
#       Copyright (C) 2009 Simon King <simon.king@nuigalway.ie> and
#                          Mike Hansen <mhansen@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
import operator
import re
from functools import reduce

from sage.rings.ring import CommutativeRing
from sage.categories.rings import Rings
from sage.structure.all import SageObject, parent
from sage.structure.factory import UniqueFactory
from sage.misc.cachefunc import cached_method

###################################################
#  The Construction Functor

from sage.categories.pushout import InfinitePolynomialFunctor


###############################################################
# Ring Factory framework

class InfinitePolynomialRingFactory(UniqueFactory):
    """
    A factory for creating infinite polynomial ring elements.  It
    makes sure that they are unique as well as handling pickling.
    For more details, see
    :class:`~sage.structure.factory.UniqueFactory` and
    :mod:`~sage.rings.polynomial.infinite_polynomial_ring`.

    EXAMPLES::

        sage: A.<a> = InfinitePolynomialRing(QQ)
        sage: B.<b> = InfinitePolynomialRing(A)
        sage: B.construction()
        [InfPoly{[a,b], "lex", "dense"}, Rational Field]
        sage: R.<a,b> = InfinitePolynomialRing(QQ)
        sage: R is B
        True
        sage: X.<x> = InfinitePolynomialRing(QQ)
        sage: X2.<x> = InfinitePolynomialRing(QQ, implementation='sparse')
        sage: X is X2
        False

        sage: X is loads(dumps(X))
        True
    """
    def create_key(self, R, names=('x',), order='lex', implementation='dense'):
        """
        Create a key which uniquely defines the infinite polynomial ring.

        TESTS::

            sage: InfinitePolynomialRing.create_key(QQ, ('y1',))
            (InfPoly{[y1], "lex", "dense"}(FractionField(...)), Integer Ring)
            sage: _[0].all
            [FractionField, InfPoly{[y1], "lex", "dense"}]
            sage: InfinitePolynomialRing.create_key(QQ, names=['beta'], order='deglex',
            ....:                                   implementation='sparse')
            (InfPoly{[beta], "deglex", "sparse"}(FractionField(...)), Integer Ring)
            sage: _[0].all
            [FractionField, InfPoly{[beta], "deglex", "sparse"}]
            sage: InfinitePolynomialRing.create_key(QQ, names=['x','y'],
            ....:                                   implementation='dense')
            (InfPoly{[x,y], "lex", "dense"}(FractionField(...)), Integer Ring)
            sage: _[0].all
            [FractionField, InfPoly{[x,y], "lex", "dense"}]

        If no generator name is provided, a generator named 'x',
        lexicographic order and the dense implementation are assumed::

            sage: InfinitePolynomialRing.create_key(QQ)
            (InfPoly{[x], "lex", "dense"}(FractionField(...)), Integer Ring)
            sage: _[0].all
            [FractionField, InfPoly{[x], "lex", "dense"}]

        If it is attempted to use no generator, a :exc:`ValueError` is raised::

            sage: InfinitePolynomialRing.create_key(ZZ, names=[])
            Traceback (most recent call last):
            ...
            ValueError: Infinite Polynomial Rings must have at least one generator
        """
        if isinstance(names, list):
            names = tuple(names)
        if not names:
            raise ValueError("Infinite Polynomial Rings must have at least one generator")
        if len(names) > len(set(names)):
            raise ValueError("the variable names must be distinct")
        F = InfinitePolynomialFunctor(names, order, implementation)
        while hasattr(R, 'construction'):
            C = R.construction()
            if C is None:
                break
            F = F * C[0]
            R = C[1]
        return (F, R)

    def create_object(self, version, key):
        """
        Return the infinite polynomial ring corresponding to the key ``key``.

        TESTS::

            sage: InfinitePolynomialRing.create_object('1.0', InfinitePolynomialRing.create_key(ZZ, ('x3',)))
            Infinite polynomial ring in x3 over Integer Ring
        """
        if len(key) > 2:
            # We got an old pickle. By calling the ring constructor, it will automatically
            # be transformed into the new scheme
            return InfinitePolynomialRing(*key)
        # By now, we have different unique keys, based on construction functors
        C, R = key
        from sage.categories.pushout import CompositeConstructionFunctor, InfinitePolynomialFunctor
        if isinstance(C, CompositeConstructionFunctor):
            F = C.all[-1]
            if len(C.all) > 1:
                R = CompositeConstructionFunctor(*C.all[:-1])(R)
        else:
            F = C
        if not isinstance(F, InfinitePolynomialFunctor):
            raise TypeError("we expected an InfinitePolynomialFunctor, not %s" % type(F))
        if F._imple == 'sparse':
            return InfinitePolynomialRing_sparse(R, F._gens, order=F._order)
        return InfinitePolynomialRing_dense(R, F._gens, order=F._order)


InfinitePolynomialRing = InfinitePolynomialRingFactory('InfinitePolynomialRing')


##############################################################
#  An auxiliary dictionary-like class that returns variables

class InfiniteGenDict:
    """
    A dictionary-like class that is suitable for usage in ``sage_eval``.

    The generators of an Infinite Polynomial Ring are not
    variables. Variables of an Infinite Polynomial Ring are returned
    by indexing a generator. The purpose of this class is to return a
    variable of an Infinite Polynomial Ring, given its string
    representation.

    EXAMPLES::

        sage: R.<a,b> = InfinitePolynomialRing(ZZ)
        sage: D = R.gens_dict() # indirect doctest
        sage: D._D
        [InfiniteGenDict defined by ['a', 'b'], {'1': 1}]
        sage: D._D[0]['a_15']
        a_15
        sage: type(_)
        <class 'sage.rings.polynomial.infinite_polynomial_element.InfinitePolynomial_dense'>
        sage: sage_eval('3*a_3*b_5-1/2*a_7', D._D[0])
        -1/2*a_7 + 3*a_3*b_5
    """
    def __init__(self, Gens):
        """
        INPUT:

        - ``Gens`` -- list of generators of an infinite polynomial ring

        EXAMPLES::

            sage: R.<a,b> = InfinitePolynomialRing(ZZ)
            sage: D = R.gens_dict() # indirect doctest
            sage: D._D
            [InfiniteGenDict defined by ['a', 'b'], {'1': 1}]
            sage: D._D == loads(dumps(D._D)) # indirect doctest
            True
        """
        self._D = dict(zip(((hasattr(X, '_name') and X._name) or repr(X) for X in Gens), Gens))

    def __eq__(self, other):
        """
        Check whether ``self`` is equal to ``other``.

        EXAMPLES::

            sage: R.<a,b> = InfinitePolynomialRing(ZZ)
            sage: D = R.gens_dict() # indirect doctest
            sage: D._D
            [InfiniteGenDict defined by ['a', 'b'], {'1': 1}]
            sage: D._D == loads(dumps(D._D)) # indirect doctest
            True
        """
        if isinstance(other, InfiniteGenDict):
            return self._D == other._D
        return False

    def __ne__(self, other):
        """
        Check whether ``self`` is not equal to ``other``.

        EXAMPLES::

            sage: R.<a,b> = InfinitePolynomialRing(ZZ)
            sage: D = R.gens_dict() # indirect doctest
            sage: D._D
            [InfiniteGenDict defined by ['a', 'b'], {'1': 1}]
            sage: D._D != loads(dumps(D._D)) # indirect doctest
            False
        """
        return not (self == other)

    def __repr__(self):
        """
        EXAMPLES::

            sage: R.<a,b> = InfinitePolynomialRing(ZZ)
            sage: D = R.gens_dict()
            sage: D._D # indirect doctest
            [InfiniteGenDict defined by ['a', 'b'], {'1': 1}]
        """
        return "InfiniteGenDict defined by %s" % sorted(self._D)

    def __getitem__(self, k):
        """
        EXAMPLES::

            sage: R.<a,b> = InfinitePolynomialRing(ZZ)
            sage: D = R.gens_dict() # indirect doctest
            sage: D._D
            [InfiniteGenDict defined by ['a', 'b'], {'1': 1}]
            sage: D._D[0]['a_15']
            a_15
            sage: type(_)
            <class 'sage.rings.polynomial.infinite_polynomial_element.InfinitePolynomial_dense'>
        """
        if not isinstance(k, str):
            raise KeyError("string expected")
        L = k.split('_')
        try:
            if len(L) == 2:
                return self._D[L[0]][int(L[1])]
        except (TypeError, ValueError):
            pass
        raise KeyError("%s is not a variable name" % k)


class GenDictWithBasering:
    """
    A dictionary-like class that is suitable for usage in ``sage_eval``.

    This pseudo-dictionary accepts strings as index, and then walks down
    a chain of base rings of (infinite) polynomial rings until it finds
    one ring that has the given string as variable name, which is then
    returned.

    EXAMPLES::

        sage: R.<a,b> = InfinitePolynomialRing(ZZ)
        sage: D = R.gens_dict() # indirect doctest
        sage: D
        GenDict of Infinite polynomial ring in a, b over Integer Ring
        sage: D['a_15']
        a_15
        sage: type(_)
        <class 'sage.rings.polynomial.infinite_polynomial_element.InfinitePolynomial_dense'>
        sage: sage_eval('3*a_3*b_5-1/2*a_7', D)
        -1/2*a_7 + 3*a_3*b_5
    """
    def __init__(self, parent, start):
        """
        INPUT:

        - ``parent`` -- a ring
        - ``start`` -- some dictionary, usually the dictionary of variables of ``parent``

        EXAMPLES::

            sage: R.<a,b> = InfinitePolynomialRing(ZZ)
            sage: D = R.gens_dict() # indirect doctest
            sage: D
            GenDict of Infinite polynomial ring in a, b over Integer Ring
            sage: D['a_15']
            a_15
            sage: type(_)
            <class 'sage.rings.polynomial.infinite_polynomial_element.InfinitePolynomial_dense'>
            sage: sage_eval('3*a_3*b_5-1/2*a_7', D)
            -1/2*a_7 + 3*a_3*b_5

        TESTS::

            sage: from sage.rings.polynomial.infinite_polynomial_ring import GenDictWithBasering
            sage: R = ZZ['x']['y']['a','b']['c']
            sage: D = GenDictWithBasering(R,R.gens_dict())
            sage: R.gens_dict()['a']
            Traceback (most recent call last):
            ...
            KeyError: 'a'
            sage: D['a']
            a
        """
        P = self._P = parent
        if isinstance(start, list):
            self._D = start
            return
        self._D = [start]
        while hasattr(P, 'base_ring') and (P.base_ring() is not P):
            P = P.base_ring()
            D = P.gens_dict()
            if isinstance(D, GenDictWithBasering):
                self._D.extend(D._D)
                break
            else:
                self._D.append(D)

    def __next__(self):
        """
        Return a dictionary that can be used to interpret strings in the base ring of ``self``.

        EXAMPLES::

            sage: R.<a,b> = InfinitePolynomialRing(QQ['t'])
            sage: D = R.gens_dict()
            sage: D
            GenDict of Infinite polynomial ring in a, b over Univariate Polynomial Ring in t over Rational Field
            sage: next(D)
            GenDict of Univariate Polynomial Ring in t over Rational Field
            sage: sage_eval('t^2', next(D))
            t^2
        """
        if len(self._D) <= 1:
            raise ValueError("no next term for %s available" % self)
        return GenDictWithBasering(self._P.base_ring(), self._D[1:])

    next = __next__

    def __repr__(self):
        """
        TESTS::

            sage: R.<a,b> = InfinitePolynomialRing(ZZ)
            sage: D = R.gens_dict() # indirect doctest
            sage: D
            GenDict of Infinite polynomial ring in a, b over Integer Ring
        """
        return "GenDict of " + repr(self._P)

    def __getitem__(self, k):
        """
        TESTS::

            sage: R.<a,b> = InfinitePolynomialRing(ZZ)
            sage: D = R.gens_dict() # indirect doctest
            sage: D
            GenDict of Infinite polynomial ring in a, b over Integer Ring
            sage: D['a_15']
            a_15
            sage: type(_)
            <class 'sage.rings.polynomial.infinite_polynomial_element.InfinitePolynomial_dense'>
        """
        for D in self._D:
            try:
                return D[k]
            except KeyError:
                pass
        raise KeyError("{} is not a variable name of {} or its iterated base rings".format(k, self._P))


##############################################################
#  The sparse implementation

class InfinitePolynomialRing_sparse(CommutativeRing):
    r"""
    Sparse implementation of Infinite Polynomial Rings.

    An Infinite Polynomial Ring with generators `x_\ast, y_\ast,
    ...` over a field `F` is a free commutative `F`-algebra generated
    by `x_0, x_1, x_2, ..., y_0, y_1, y_2, ..., ...` and is equipped
    with a permutation action on the generators, namely `x_n^P =
    x_{P(n)}, y_{n}^P=y_{P(n)}, ...` for any permutation `P` (note
    that variables of index zero are invariant under such
    permutation).

    It is known that any permutation invariant ideal in an Infinite
    Polynomial Ring is finitely generated modulo the permutation
    action -- see :class:`~sage.rings.polynomial.symmetric_ideal.SymmetricIdeal`
    for more details.

    Usually, an instance of this class is created using
    ``InfinitePolynomialRing`` with the optional parameter
    ``implementation='sparse'``. This takes care of uniqueness of
    parent structures. However, a direct construction is possible, in
    principle::

        sage: X.<x,y> = InfinitePolynomialRing(QQ, implementation='sparse')
        sage: Y.<x,y> = InfinitePolynomialRing(QQ, implementation='sparse')
        sage: X is Y
        True
        sage: from sage.rings.polynomial.infinite_polynomial_ring import InfinitePolynomialRing_sparse
        sage: Z = InfinitePolynomialRing_sparse(QQ, ['x','y'], 'lex')

    Nevertheless, since infinite polynomial rings are supposed to be unique
    parent structures, they do not evaluate equal. ::

        sage: Z == X
        False

    The last parameter ('lex' in the above example) can also be
    'deglex' or 'degrevlex'; this would result in an Infinite
    Polynomial Ring in degree lexicographic or degree reverse
    lexicographic order.

    See :mod:`~sage.rings.polynomial.infinite_polynomial_ring` for
    more details.
    """
    def __init__(self, R, names, order):
        """
        INPUT:

        - ``R`` -- base ring
        - ``names`` -- list of generator names
        - ``order`` -- string determining the monomial order of the infinite polynomial ring

        EXAMPLES::

            sage: X.<alpha,beta> = InfinitePolynomialRing(ZZ, implementation='sparse')

        Infinite Polynomial Rings are unique parent structures::

            sage: X is loads(dumps(X))
            True
            sage: p=alpha[10]*beta[2]^3+2*alpha[1]*beta[3]
            sage: p
            alpha_10*beta_2^3 + 2*alpha_1*beta_3

        We define another Infinite Polynomial Ring with same generator
        names but a different order. These rings are different, but
        allow for coercion::

            sage: Y.<alpha,beta> = InfinitePolynomialRing(QQ, order='deglex', implementation='sparse')
            sage: Y is X
            False
            sage: q=beta[2]^3*alpha[10]+beta[3]*alpha[1]*2
            sage: q
            alpha_10*beta_2^3 + 2*alpha_1*beta_3
            sage: p==q
            True
            sage: X.gen(1)[2]*Y.gen(0)[1]
            alpha_1*beta_2
        """
        if not names:
            names = ['x']
        for n in names:
            if not (isinstance(n, str) and n.isalnum() and (not n[0].isdigit())):
                raise ValueError("generator names must be alphanumeric strings not starting with a digit, but %s is not" % n)
        if len(names) != len(set(names)):
            raise ValueError("generator names must be pairwise different")
        self._names = tuple(names)
        if not isinstance(order, str):
            raise TypeError("the monomial order must be given as a string")
        if R not in Rings().Commutative():
            raise TypeError("the given 'base ring' (= %s) must be a commutative ring" % R)

        # now, the input is accepted
        if hasattr(R, '_underlying_ring'):
            self._underlying_ring = R._underlying_ring
        else:
            self._underlying_ring = R.base_ring()

        # some tools to analyse polynomial string representations.
        self._identify_variable = lambda x, y: (-self._names.index(x), int(y))
        self._find_maxshift = re.compile('_([0-9]+)')  # findall yields stringrep of the shifts
        self._find_variables = re.compile('[a-zA-Z0-9]+_[0-9]+')
        self._find_varpowers = re.compile(r'([a-zA-Z0-9]+)_([0-9]+)\^?([0-9]*)')  # findall yields triple "generator_name", "index", "exponent"

        # Create some small underlying polynomial ring.
        # It is used to ensure that the parent of the underlying
        # polynomial of an element of self is actually a *multi*variate
        # polynomial ring.
        from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
        if len(names) == 1:
            VarList = [names[0] + '_0', names[0] + '_1']
        else:
            VarList = [X + '_0' for X in names]
        VarList.sort(key=self.varname_key, reverse=True)
        self._minP = PolynomialRing(R, len(VarList), VarList)

        # some basic data
        self._order = order
        self._name_dict = {name: i for i, name in enumerate(names)}
        from sage.categories.commutative_algebras import CommutativeAlgebras
        CommutativeRing.__init__(self, R, category=CommutativeAlgebras(R))

        self._populate_coercion_lists_()

    def __repr__(self):
        """
        EXAMPLES::

            sage: InfinitePolynomialRing(QQ)         # indirect doctest
            Infinite polynomial ring in x over Rational Field

            sage: X.<alpha,beta> = InfinitePolynomialRing(ZZ, order='deglex'); X
            Infinite polynomial ring in alpha, beta over Integer Ring
        """
        return "Infinite polynomial ring in {} over {}".format(", ".join(self._names), self._base)

    def _latex_(self):
        r"""
        EXAMPLES::

            sage: from sage.misc.latex import latex
            sage: X.<x,y> = InfinitePolynomialRing(QQ)
            sage: latex(X)        # indirect doctest
            \Bold{Q}[x_{\ast}, y_{\ast}]
        """
        from sage.misc.latex import latex
        vars = ', '.join(latex(X) for X in self.gens())
        return f"{latex(self.base_ring())}[{vars}]"

    @cached_method
    def _an_element_(self):
        """
        Return an element of this ring.

        EXAMPLES::

            sage: R.<x> = InfinitePolynomialRing(QQ)
            sage: R.an_element() # indirect doctest
            x_1
        """
        x = self.gen(0)
        return x[1]

    @cached_method
    def one(self):
        """
        TESTS::

            sage: X.<x,y> = InfinitePolynomialRing(QQ)
            sage: X.one()
            1
        """
        from sage.rings.polynomial.infinite_polynomial_element import InfinitePolynomial
        return InfinitePolynomial(self, self._base(1))

    #####################
    # coercion

    def construction(self):
        """
        Return the construction of ``self``.

        OUTPUT:

        A pair ``F,R``, where ``F`` is a construction functor and ``R`` is a ring,
        so that ``F(R) is self``.

        EXAMPLES::

            sage: R.<x,y> = InfinitePolynomialRing(GF(5))
            sage: R.construction()
            [InfPoly{[x,y], "lex", "dense"}, Finite Field of size 5]
        """
        return [InfinitePolynomialFunctor(self._names, self._order, 'sparse'), self._base]

    def _coerce_map_from_(self, S):
        r"""
        Coerce things into ``self``.

        NOTE:

        Any coercion will preserve variable names.

        EXAMPLES:

        Here, we check to see that elements of a *finitely* generated
        polynomial ring with appropriate variable names coerce
        correctly into the Infinite Polynomial Ring::

            sage: X.<x> = InfinitePolynomialRing(QQ)
            sage: px0 = PolynomialRing(QQ,'x_0').gen(0)
            sage: px0 + x[0]  # indirect doctest
            2*x_0
            sage: px0==x[0]
            True

        It is possible to construct an Infinite Polynomial Ring whose
        base ring is another Infinite Polynomial Ring::

            sage: R.<a,b> = InfinitePolynomialRing(ZZ)
            sage: X.<x> = InfinitePolynomialRing(R)
            sage: a[2]*x[3]+x[1]*a[4]^2
            a_4^2*x_1 + a_2*x_3
        """
        # Use Construction Functors!
        from sage.categories.pushout import pushout
        try:
            # the following line should not test "pushout is self", but
            # only "pushout == self", since we also allow coercion from
            # dense to sparse implementation!
            P = pushout(self, S)
            # We do not care about the orders. But base ring and generators
            # of the pushout should remain the same as in self.
            return P._names == self._names and P._base == self._base
        except (TypeError, ValueError):
            return False

    def _element_constructor_(self, x):
        """
        Return an element of ``self``.

        INPUT:

        - ``x`` -- any object that can be interpreted in ``self``

        TESTS::

            sage: X = InfinitePolynomialRing(QQ)
            sage: a = X(2); a     # indirect doctest
            2
            sage: a.parent()
            Infinite polynomial ring in x over Rational Field

            sage: R = PolynomialRing(ZZ, ['x_3'])
            sage: b = X(R.gen()); b
            x_3
            sage: b.parent()
            Infinite polynomial ring in x over Rational Field
            sage: X('(x_1^2+2/3*x_4)*(x_2+x_5)')
            2/3*x_5*x_4 + x_5*x_1^2 + 2/3*x_4*x_2 + x_2*x_1^2

            sage: Y = InfinitePolynomialRing(ZZ)
            sage: Y('1/3')
            Traceback (most recent call last):
            ...
            ValueError: cannot convert 1/3 into an element of Infinite polynomial ring in x over Integer Ring

        .. WARNING::

            The :issue:`37756` is not yet fixed::

                sage: L.<x, y> = QQ[]
                sage: R.<a> = InfinitePolynomialRing(QQ)
                sage: M = InfinitePolynomialRing(L, names=["a"])
                sage: c = a[0]
                sage: M(c)  # known bug
                a_0
        """
        from sage.rings.polynomial.infinite_polynomial_element import InfinitePolynomial
        # In many cases, the easiest solution is to "simply" evaluate
        # the string representation.
        from sage.misc.sage_eval import sage_eval
        if isinstance(x, str):
            try:
                x = sage_eval(x, self.gens_dict())
            except (TypeError, ValueError, SyntaxError):
                raise ValueError(f"cannot convert {x} into an element of {self}")
            P = parent(x)
            if P is self:
                return x
            elif self._base.has_coerce_map_from(P):
                return InfinitePolynomial(self, self._base(x))
            else:
                raise ValueError(f"cannot convert {x} into an element of {self}")

        if isinstance(parent(x), InfinitePolynomialRing_sparse):
            # the easy case - parent == self - is already past
            if x.parent() is self._base:  # another easy case
                return InfinitePolynomial(self, x)
            xmaxind = x.max_index()  # save for later
            x = x._p
        else:
            xmaxind = -1

        # Now, we focus on the underlying classical polynomial ring.
        # First, try interpretation in the base ring.
        try:
            from sage.rings.polynomial.multi_polynomial_ring import MPolynomialRing_polydict
            if isinstance(self._base, MPolynomialRing_polydict):
                x = sage_eval(repr(), next(self.gens_dict()))
            else:
                x = self._base(x)
            # remark: Conversion to self._P (if applicable)
            # is done in InfinitePolynomial()
            return InfinitePolynomial(self, x)
        except (TypeError, ValueError):
            pass

        # By now, we can assume that x has a parent, because
        # types like int have already been done in the previous step;
        # and also it is not an InfinitePolynomial.
        # If it is not a polynomial (duck typing: we need
        # the variables attribute), we fall back to using strings
        if not hasattr(x, 'variables'):
            try:
                return sage_eval(repr(x), self.gens_dict())
            except (TypeError, ValueError, SyntaxError):
                raise ValueError(f"cannot convert {x} into an element of {self}")

        # direct conversion will only be used if the underlying polynomials are libsingular.
        from sage.rings.polynomial.multi_polynomial import MPolynomial_libsingular
        # try interpretation in self._P, if we have a dense implementation
        if hasattr(self, '_P'):
            if x.parent() is self._P:
                return InfinitePolynomial(self, x)
            # It's a shame to use sage_eval. However, it's even more of a shame
            # that MPolynomialRing_polydict does not work in complicated settings.
            # So, if self._P is libsingular (and this will be the case in many
            # applications!), we do it "nicely". Otherwise, we have to use sage_eval.
            if isinstance(x, MPolynomial_libsingular):
                from sage.rings.polynomial.multi_polynomial_libsingular import MPolynomialRing_libsingular
                if isinstance(self._P, MPolynomialRing_libsingular):
                    if xmaxind == -1:  # Otherwise, x has been an InfinitePolynomial
                        # We infer the correct variable shift.
                        # Note: Since we are in the "libsingular" case, there are
                        # no further "variables" hidden in the base ring of x.parent()
                        try:
                            VarList = [repr(v) for v in x.variables()]
                            # since interpretation in base ring
                            # was impossible, it *must* have
                            # variables
                            # This tests admissibility on the fly:
                            VarList.sort(key=self.varname_key, reverse=True)
                        except ValueError:
                            raise ValueError("cannot convert {} into an element of {} - variables are not admissible".format(x, self))
                        xmaxind = max([int(v.split('_')[1]) for v in VarList])
                    try:
                        # Apparently, in libsingular, the polynomial conversion is not done by
                        # name but by position, if the number of variables in the parents coincide.
                        # So, we shift self._P to achieve xmaxind, and if the number of variables is
                        # the same then we shift further. We then *must* be
                        # able to convert x into self._P, or conversion to self is
                        # impossible (and will be done in InfinitePolynomial(...)
                        if self._max < xmaxind:
                            self.gen()[xmaxind]
                        if self._P.ngens() == x.parent().ngens():
                            self.gen()[self._max + 1]
                        # conversion to self._P will be done in InfinitePolynomial.__init__
                        return InfinitePolynomial(self, x)
                    except (ValueError, TypeError, NameError):
                        raise ValueError("cannot convert {} (from {}, but variables {}) into an element of {} - no conversion into underlying polynomial ring {}".format(x, x.parent(), x.variables(), self, self._P))
            # By now, x or self._P are not libsingular. Since MPolynomialRing_polydict
            # is too buggy, we use string evaluation
            try:
                return sage_eval(repr(x), self.gens_dict())
            except (ValueError, TypeError, NameError):
                raise ValueError("cannot convert {} into an element of {} - no conversion into underlying polynomial ring".format(x, self))

        # By now, we are in the sparse case.
        try:
            VarList = [repr(v) for v in x.variables()]
            # since interpretation in base ring
            # was impossible, it *must* have
            # variables
            # This tests admissibility on the fly:
            VarList.sort(key=self.varname_key, reverse=True)
        except ValueError:
            raise ValueError("cannot convert {} into an element of {} - variables are not admissible".format(x, self))

        if len(VarList) == 1:
            # univariate polynomial rings are crab. So, make up another variable.
            if VarList[0] == self._names[0] + '_0':
                VarList.append(self._names[0] + '_1')
            else:
                VarList.append(self._names[0] + '_0')
        # We ensure that polynomial conversion is done by names;
        # the problem is that it is done by names if the number of variables coincides.
        if len(VarList) == x.parent().ngens():
            BigList = x.parent().variable_names()
            ind = 2
            while self._names[0] + '_' + str(ind) in BigList:
                ind += 1
            VarList.append(self._names[0] + '_' + str(ind))
        try:
            VarList.sort(key=self.varname_key, reverse=True)
        except ValueError:
            raise ValueError("cannot convert {} into an element of {}; the variables are not admissible".format(x, self))

        from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
        R = PolynomialRing(self._base, VarList, order=self._order)
        if isinstance(x, MPolynomial_libsingular):  # everything else is so buggy that it's even not worth to try.
            from sage.rings.polynomial.multi_polynomial_libsingular import MPolynomialRing_libsingular
            if isinstance(R, MPolynomialRing_libsingular):
                try:
                    # Problem: If there is only a partial overlap in the variables
                    # of x.parent() and R, then R(x) raises an error (which, I think,
                    # is a bug, since we talk here about conversion, not coercion).
                    # Hence, for being on the safe side, we coerce into a pushout ring:
                    x = R(1) * x
                    return InfinitePolynomial(self, x)
                except (TypeError, ValueError):
                    # OK, last resort, to be on the safe side
                    try:
                        return sage_eval(repr(x), self.gens_dict())
                    except (ValueError, TypeError, NameError):
                        raise ValueError("cannot convert {} into an element of {}; conversion of the underlying polynomial failed".format(x, self))
        try:
            return sage_eval(repr(x), self.gens_dict())
        except (ValueError, TypeError, NameError):
            raise ValueError(f"cannot convert {x} into an element of {self}")

    def tensor_with_ring(self, R):
        """
        Return the tensor product of ``self`` with another ring.

        INPUT:

        - ``R`` -- a ring

        OUTPUT:

        An infinite polynomial ring that, mathematically, can be seen as the
        tensor product of ``self`` with ``R``.

        NOTE:

        It is required that the underlying ring of ``self`` coerces into ``R``.
        Hence, the tensor product is in fact merely an extension of the base
        ring.

        EXAMPLES::

            sage: R.<a,b> = InfinitePolynomialRing(ZZ)
            sage: R.tensor_with_ring(QQ)
            Infinite polynomial ring in a, b over Rational Field
            sage: R
            Infinite polynomial ring in a, b over Integer Ring

        The following tests against a bug that was fixed at :issue:`10468`::

            sage: R.<x,y> = InfinitePolynomialRing(QQ)
            sage: R.tensor_with_ring(QQ) is R
            True
        """
        if not R.has_coerce_map_from(self._underlying_ring):
            raise TypeError("we cannot tensor with " + repr(R))
        B = self.base_ring()
        if hasattr(B, 'tensor_with_ring'):
            return InfinitePolynomialRing(B.tensor_with_ring(R), self._names, self._order, implementation='sparse')
        if hasattr(B, 'change_ring'):  # e.g., polynomial rings
            return InfinitePolynomialRing(B.change_ring(R), self._names, self._order, implementation='sparse')
        # try to find the correct base ring in other ways:
        try:
            o = B.one() * R.one()
        except (TypeError, ValueError):
            raise TypeError("we cannot tensor with " + repr(R))
        return InfinitePolynomialRing(o.parent(), self._names, self._order, implementation='sparse')

    # Basic Ring Properties
    # -- some stuff that is useful for quotient rings etc.
    def is_noetherian(self):
        """
        Return ``False``, since polynomial rings in infinitely many
        variables are never Noetherian rings.

        Since Infinite Polynomial Rings must have at least one
        generator, they have infinitely many variables and are thus
        not Noetherian, as a ring.

        .. NOTE::

            Infinite Polynomial Rings over a field `F` are Noetherian as
            `F(G)` modules, where `G` is the symmetric group of the
            natural numbers. But this is not what the method
            ``is_noetherian()`` is answering.

        TESTS::

            sage: R = InfinitePolynomialRing(GF(2))
            sage: R
            Infinite polynomial ring in x over Finite Field of size 2
            sage: R.is_noetherian()
            False

            sage: R.<x> = InfinitePolynomialRing(QQ)
            sage: R.is_noetherian()
            False
        """
        return False

    def is_field(self, *args, **kwds):
        """
        Return ``False`` since Infinite Polynomial Rings are never fields.

        Since Infinite Polynomial Rings must have at least one generator,
        they have infinitely many variables and thus never are fields.

        EXAMPLES::

            sage: R.<x, y> = InfinitePolynomialRing(QQ)
            sage: R.is_field()
            False

        TESTS::

            sage: R = InfinitePolynomialRing(GF(2))
            sage: R
            Infinite polynomial ring in x over Finite Field of size 2
            sage: R.is_field()
            False

        :issue:`9443`::

            sage: W = PowerSeriesRing(InfinitePolynomialRing(QQ,'a'),'x')
            sage: W.is_field()
            False
        """
        return False

    def varname_key(self, x):
        """
        Key for comparison of variable names.

        INPUT:

        - ``x`` -- string of the form ``a+'_'+str(n)``, where a is the
          name of a generator, and n is an integer

        OUTPUT: a key used to sort the variables

        THEORY:

        The order is defined as follows:

        x<y `\\iff` the string ``x.split('_')[0]`` is later in the list of
        generator names of ``self`` than ``y.split('_')[0]``, or
        (``x.split('_')[0]==y.split('_')[0]`` and
        ``int(x.split('_')[1])<int(y.split('_')[1])``)

        EXAMPLES::

            sage: X.<alpha,beta> = InfinitePolynomialRing(ZZ)
            sage: X.varname_key('alpha_1')
            (0, 1)
            sage: X.varname_key('beta_10')
            (-1, 10)
            sage: X.varname_key('beta_1')
            (-1, 1)
            sage: X.varname_key('alpha_10')
            (0, 10)
            sage: X.varname_key('alpha_1')
            (0, 1)
            sage: X.varname_key('alpha_10')
            (0, 10)
        """
        try:
            return self._identify_variable(*x.split('_', 1))
        except (KeyError, ValueError, TypeError):
            raise ValueError("%s is not a valid variable name" % x)

    def ngens(self):
        """
        Return the number of generators for this ring.

        Since there
        are countably infinitely many variables in this polynomial
        ring, by 'generators' we mean the number of infinite families
        of variables. See :mod:`~sage.rings.polynomial.infinite_polynomial_ring`
        for more details.

        EXAMPLES::

            sage: X.<x> = InfinitePolynomialRing(ZZ)
            sage: X.ngens()
            1

            sage: X.<x1,x2> = InfinitePolynomialRing(QQ)
            sage: X.ngens()
            2
        """
        return len(self._names)

    @cached_method
    def gen(self, i=None):
        """
        Return the `i`-th 'generator' (see the description in :meth:`.ngens`)
        of this infinite polynomial ring.

        EXAMPLES::

            sage: X = InfinitePolynomialRing(QQ)
            sage: x = X.gen()
            sage: x[1]
            x_1
            sage: X.gen() is X.gen(0)
            True
            sage: XX = InfinitePolynomialRing(GF(5))
            sage: XX.gen(0) is XX.gen()
            True
        """
        if i is not None and i > len(self._names):
            raise ValueError
        j = i if i is not None else 0
        res = InfinitePolynomialGen(self, self._names[j])
        if i is None:
            key = ((0,), ())
            if key in self._cache__gen:
                return self._cache__gen[key]
            else:
                self._cache__gen[key] = res
        return res

    def _first_ngens(self, n):
        """
        Used by the preparser for R.<x> = ...

        EXAMPLES::

            sage: InfinitePolynomialRing(ZZ, 'a')._first_ngens(1)
            (a_*,)
        """
        # It may be that we merge variables. If this is the case,
        # the new variables (as used by R.<x>  = ...) come *last*,
        # but in order.
        return self.gens()[-n:]

    @cached_method
    def gens_dict(self):
        """
        Return a dictionary-like object containing the infinitely many
        ``{var_name:variable}`` pairs.

        EXAMPLES::

            sage: R = InfinitePolynomialRing(ZZ, 'a')
            sage: D = R.gens_dict()
            sage: D
            GenDict of Infinite polynomial ring in a over Integer Ring
            sage: D['a_5']
            a_5
        """
        return GenDictWithBasering(self, InfiniteGenDict(self.gens()))

    def _ideal_class_(self, n=0):
        """
        Return :class:`SymmetricIdeals` (see there for further details).

        TESTS::

            sage: R.<a,b> = InfinitePolynomialRing(ZZ)
            sage: R._ideal_class_()
            <class 'sage.rings.polynomial.symmetric_ideal.SymmetricIdeal'>
        """
        import sage.rings.polynomial.symmetric_ideal
        return sage.rings.polynomial.symmetric_ideal.SymmetricIdeal

    def characteristic(self):
        """
        Return the characteristic of the base field.

        EXAMPLES::

            sage: X.<x,y> = InfinitePolynomialRing(GF(25,'a'))                          # needs sage.rings.finite_rings
            sage: X                                                                     # needs sage.rings.finite_rings
            Infinite polynomial ring in x, y over Finite Field in a of size 5^2
            sage: X.characteristic()                                                    # needs sage.rings.finite_rings
            5
        """
        return self._base.characteristic()

    def is_integral_domain(self, *args, **kwds):
        """
        An infinite polynomial ring is an integral domain if and only if
        the base ring is.  Arguments are passed to is_integral_domain
        method of base ring.

        EXAMPLES::

            sage: R.<x, y> = InfinitePolynomialRing(QQ)
            sage: R.is_integral_domain()
            True

        TESTS:

        :issue:`9443`::

            sage: W = PolynomialRing(InfinitePolynomialRing(QQ,'a'),2,'x,y')
            sage: W.is_integral_domain()
            True
        """
        return self._base.is_integral_domain(*args, **kwds)

    def krull_dimension(self, *args, **kwds):
        """
        Return ``Infinity``, since polynomial rings in infinitely many
        variables have infinite Krull dimension.

        EXAMPLES::

            sage: R.<x, y> = InfinitePolynomialRing(QQ)
            sage: R.krull_dimension()
            +Infinity
        """
        from sage.rings.infinity import Infinity
        return Infinity

    def order(self):
        """
        Return ``Infinity``, since polynomial rings have infinitely
        many elements.

        EXAMPLES::

            sage: R.<x> = InfinitePolynomialRing(GF(2))
            sage: R.order()
            +Infinity
        """
        from sage.rings.infinity import Infinity
        return Infinity

    # Other bases
    def key_basis(self):
        r"""
        Return the basis of ``self`` given by key polynomials.

        EXAMPLES::

            sage: R.<x> = InfinitePolynomialRing(GF(2))
            sage: R.key_basis()                                                         # needs sage.combinat sage.modules
            Key polynomial basis over Finite Field of size 2
        """
        from sage.combinat.key_polynomial import KeyPolynomialBasis
        return KeyPolynomialBasis(self)


class InfinitePolynomialGen(SageObject):
    """
    This class provides the object which is responsible for returning
    variables in an infinite polynomial ring (implemented in
    :meth:`.__getitem__`).

    EXAMPLES::

        sage: # needs sage.rings.real_mpfr
        sage: X.<x1,x2> = InfinitePolynomialRing(RR)
        sage: x1
        x1_*
        sage: x1[5]
        x1_5
        sage: x1 == loads(dumps(x1))
        True
    """

    def __init__(self, parent, name):
        """
        EXAMPLES::

            sage: X.<x> = InfinitePolynomialRing(QQ)
            sage: loads(dumps(x))
            x_*
        """
        self._name = name
        self._parent = parent
        self._output = {}

    def __eq__(self, other):
        """
        Check whether ``self`` is equal to ``other``.

        EXAMPLES::

            sage: X.<x,y> = InfinitePolynomialRing(QQ)
            sage: from sage.rings.polynomial.infinite_polynomial_ring import InfinitePolynomialGen
            sage: x2 = InfinitePolynomialGen(X, 'x')
            sage: x2 == x
            True
        """
        if not isinstance(other, InfinitePolynomialGen):
            return False
        return (self._name, self._parent) == (other._name, other._parent)

    def __ne__(self, other):
        """
        Check whether ``self`` is not equal to ``other``.

        EXAMPLES::

            sage: X.<x,y> = InfinitePolynomialRing(QQ)
            sage: from sage.rings.polynomial.infinite_polynomial_ring import InfinitePolynomialGen
            sage: x2 = InfinitePolynomialGen(X, 'x')
            sage: x2 != x
            False
        """
        return not (self == other)

    def _latex_(self):
        r"""
        EXAMPLES::

            sage: from sage.misc.latex import latex
            sage: X.<x,x1,xx> = InfinitePolynomialRing(QQ)
            sage: latex(x) # indirect doctest
            x_{\ast}
            sage: latex(x1) # indirect doctest
            \mathit{x1}_{\ast}
            sage: latex(xx) # indirect doctest
            \mathit{xx}_{\ast}
            sage: latex(x[2]) # indirect doctest
            x_{2}
            sage: latex(x1[3]) # indirect doctest
            \mathit{x1}_{3}
        """
        from sage.misc.latex import latex_variable_name
        return latex_variable_name(self._name + '_ast')

    def __getitem__(self, i):
        """
        Return the variable ``x[i]`` where ``x`` is this
        :class:`sage.rings.polynomial.infinite_polynomial_ring.InfinitePolynomialGen`,
        and i is a nonnegative integer.

        EXAMPLES::

            sage: X.<alpha> = InfinitePolynomialRing(QQ)
            sage: alpha[1]
            alpha_1
        """
        if int(i) != i:
            raise ValueError("the index (= %s) must be an integer" % i)
        i = int(i)
        if i < 0:
            raise ValueError("the index (= %s) must be nonnegative" % i)
        P = self._parent
        from sage.rings.polynomial.infinite_polynomial_element import InfinitePolynomial_dense, InfinitePolynomial_sparse
        from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
        OUT = self._output.get(i)
        if hasattr(P, '_P'):
            if i <= P._max:
                # return InfinitePolynomial_dense(P, P._P.gen(P._P.variable_names().index(self._name+'_'+str(i))))
                if OUT is None:
                    self._output[i] = InfinitePolynomial_dense(P, P._P.gen(P._P.variable_names().index(self._name + '_' + str(i))))
                else:
                    if OUT._p.parent() is not P._P:
                        OUT._p = P._P(OUT._p)
                return self._output[i]
            # Calculate all of the new names needed
            try:
                names = [[name + '_' + str(j) for name in P._names]
                         for j in range(i + 1)]
            except OverflowError:
                raise IndexError("variable index is too big - consider using the sparse implementation")
            names = reduce(operator.add, names)
            names.sort(key=P.varname_key, reverse=True)
            # Create the new polynomial ring
            P._P = PolynomialRing(P.base_ring(), names, order=P._order)
            # Get the generators
            P._max = i
            # return InfinitePolynomial_dense(P, P._P.gen(P._P.variable_names().index(self._name+'_'+str(i))))
            self._output[i] = InfinitePolynomial_dense(P, P._P.gen(P._P.variable_names().index(self._name + '_' + str(i))))
            return self._output[i]
        # Now, we are in the sparse implementation
        if OUT is not None:  # in the sparse implementation, this is ok
            return OUT
        if i == 0:
            names = [self._name + '_0', self._name + '_1']
        else:
            names = [self._name + '_0', self._name + '_' + str(i)]
        names.sort(key=P.varname_key, reverse=True)
        Pol = PolynomialRing(P.base_ring(), names, order=P._order)
        # return InfinitePolynomial_sparse(P, Pol.gen(names.index(self._name+'_'+str(i))))
        self._output[i] = InfinitePolynomial_sparse(P, Pol.gen(names.index(self._name + '_' + str(i))))
        return self._output[i]

    def _repr_(self):
        """
        EXAMPLES::

            sage: X.<x,y> = InfinitePolynomialRing(QQ)
            sage: x  # indirect doctest
            x_*
        """
        return self._name + '_*'

    def __str__(self):
        """
        EXAMPLES::

            sage: X.<x,y> = InfinitePolynomialRing(QQ)
            sage: print(x) # indirect doctest
            Generator for the x's in Infinite polynomial ring in x, y over Rational Field
        """
        return f"Generator for the {self._name}'s in {self._parent}"


##############################################################
#  The dense implementation

class InfinitePolynomialRing_dense(InfinitePolynomialRing_sparse):
    """
    Dense implementation of Infinite Polynomial Rings.

    Compared with :class:`~sage.rings.polynomial.infinite_polynomial_ring.InfinitePolynomialRing_sparse`,
    from which this class inherits, it keeps a polynomial ring that comprises all elements that have
    been created so far.
    """
    def __init__(self, R, names, order):
        """
        EXAMPLES::

            sage: X.<x2,alpha,y4> = InfinitePolynomialRing(ZZ, implementation='dense')
            sage: X == loads(dumps(X))
            True
        """
        if not names:
            names = ['x']
        # Generate the initial polynomial ring
        self._max = 0
        InfinitePolynomialRing_sparse.__init__(self, R, names, order)
        self._P = self._minP
        # self._pgens = self._P.gens()

    #####################
    # Coercion

    def construction(self):
        """
        Return the construction of ``self``.

        OUTPUT:

        A pair ``F,R``, where ``F`` is a construction functor and ``R`` is a ring,
        so that ``F(R) is self``.

        EXAMPLES::

            sage: R.<x,y> = InfinitePolynomialRing(GF(5))
            sage: R.construction()
            [InfPoly{[x,y], "lex", "dense"}, Finite Field of size 5]
        """
        return [InfinitePolynomialFunctor(self._names, self._order, 'dense'), self._base]

    def tensor_with_ring(self, R):
        """
        Return the tensor product of ``self`` with another ring.

        INPUT:

        - ``R`` -- a ring

        OUTPUT:

        An infinite polynomial ring that, mathematically, can be seen as the
        tensor product of ``self`` with ``R``.

        NOTE:

        It is required that the underlying ring of ``self`` coerces into ``R``.
        Hence, the tensor product is in fact merely an extension of the base
        ring.

        EXAMPLES::

            sage: R.<a,b> = InfinitePolynomialRing(ZZ, implementation='sparse')
            sage: R.tensor_with_ring(QQ)
            Infinite polynomial ring in a, b over Rational Field
            sage: R
            Infinite polynomial ring in a, b over Integer Ring

        The following tests against a bug that was fixed at :issue:`10468`::

            sage: R.<x,y> = InfinitePolynomialRing(QQ, implementation='sparse')
            sage: R.tensor_with_ring(QQ) is R
            True
        """
        if not R.has_coerce_map_from(self._underlying_ring):
            raise TypeError("we cannot tensor with " + repr(R))
        B = self.base_ring()
        if hasattr(B, 'tensor_with_ring'):
            return InfinitePolynomialRing(B.tensor_with_ring(R), self._names, self._order, implementation='dense')
        if hasattr(B, 'change_ring'):  # e.g., polynomial rings
            return InfinitePolynomialRing(B.change_ring(R), self._names, self._order, implementation='dense')
        # try to find the correct base ring in other ways:
        try:
            o = B.one() * R.one()
        except (TypeError, ValueError):
            raise TypeError("we cannot tensor with " + repr(R))
        return InfinitePolynomialRing(o.parent(), self._names, self._order, implementation='dense')

    def polynomial_ring(self):
        """
        Return the underlying *finite* polynomial ring.

        .. NOTE::

           The ring returned can change over time as more variables
           are used.

           Since the rings are cached, we create here a ring with variable
           names that do not occur in other doc tests, so that we avoid
           side effects.

        EXAMPLES::

            sage: X.<xx, yy> = InfinitePolynomialRing(ZZ)
            sage: X.polynomial_ring()
            Multivariate Polynomial Ring in xx_0, yy_0 over Integer Ring
            sage: a = yy[3]
            sage: X.polynomial_ring()
            Multivariate Polynomial Ring in xx_3, xx_2, xx_1, xx_0, yy_3, yy_2, yy_1, yy_0
             over Integer Ring
        """
        return self._P
